#!/usr/bin/env python

"""This script produces a measure of the irregularity of
a waveform.  It is zero if the waveform is precisely periodic.
It is approximately 1 for white noise.

This is used as the aperiodicity measure in
Kochanski, Coleman, Grabe and Rosner JASA 2005.
"""

import math
import Num
import gpkimgclass
import power
import die
import voice_misc

DT = 0.010	# Output sampling interval.
DELTA_DELAY = 1.0/(1200.0 * 2 * math.pi)
# DELTA_DELAY is chosen so that we can align the oscillations
# at the first formant (f1 is normally under 1200 Hz)
# to an accuracy of 1 radian or less.

Fmin = 50.0
Fmax = 500.0
Fhp = 500.0
WINDOW = 0.020

DBG = None

def one_predict(d, delay, dt, Dt):
	zl = Num.zeros((delay/2,), Num.Float)
	zr = Num.zeros((delay-delay/2,), Num.Float)
	dtmp = Num.concatenate( (zl, d, zr) )
	tmp = dtmp[delay:] - dtmp[:-delay]
	assert tmp.shape == d.shape
	return power.smooth(tmp*tmp, dt, Dt, extra=WINDOW)

def predict_err(d, delays, dt, Dt):
	o = None
	for delay in delays:
		tmp,t0 = one_predict(d, delay, dt, Dt)
		if o is None:
			o = tmp
		else:
			Num.minimum(o, tmp, o)
	return o


def local_power(d, dt, Dt):
	return power.smooth(d*d, dt, Dt, extra=WINDOW)


def fractional_step_range(initial, final, step=1):
	"""Generates an array of window offsets.
	No two offsets are equal, and they are integers
	between initial and final."""

	assert step != 0
	if step < 0:
		return fractional_step_range(initial, final, -step)
	if final < initial:
		return []
	o = []
	tmp = float(initial)
	assert tmp > 0.5, "Initial=%g too close to zero, or even negative!" % tmp
	ilast = 0
	while tmp <= final:
		itmp = int(round(tmp))
		if itmp != ilast and itmp>=initial:
			o.append(itmp)
		ilast = itmp
		tmp += step
	return o


def  steady_state(d, dt, Dt):
	if Fmin*dt > 0.5 or Fmin*dt<1.0/d.shape[0]:
		die.warn("Silly time scale")
	die.info("Hipass at %f" % (Fmin*dt))
	d = voice_misc.hipass_sym_butterworth(d, Fmin*dt)
	d = voice_misc.hipass_first_order(d, Fhp*dt)
	if DBG:
		gpkimgclass.gpk_img({'CDELT2': dt}, d).write(DBG)
	p,tshift = local_power(d, dt, Dt)
	assert tshift < Dt
	imin = 1.0/(dt * Fmax)
	imax = 1.0/(dt * Fmin)
	istep = DELTA_DELAY/dt
	delays = fractional_step_range(imin, imax, istep)
	perr = predict_err(d, delays, dt, Dt)
		# The factor of 2.0 below is because perr is a difference
		# of two signals.  If they were independent random noise,
		# then the variances would add, so that perr=2*p for
		# white noise.
		# In other words, this measure goes to 1 for white noise,
		# and to zero for a periodic signal.
	assert perr.shape == p.shape
	return Num.sqrt(perr/(2.0*p))




if __name__ == '__main__':
	import sys

	try:
		import psyco
		psyco.full(memory=10000)
	except ImportError:
		pass

	arglist = sys.argv[1:]
	fname = None
	column = None
	out = "foo.irr.d"
	extrahdr = {}
	while len(arglist)>0 and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-f':
			fname = arglist.pop(0)
		elif arg == '-c':
			column = arglist.pop(0)
			try:
				column = int(column)
			except ValueError:
				pass
		elif arg == '-d':
			DBG = arglist.pop(0)
		elif arg == '-o':
			out = arglist.pop(0)
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.info("Unrecognized flag: %s" % arg)
			print __doc__
			die.exit(1)
	d = gpkimgclass.read(fname)
	ss = steady_state(d.column(column), d.dt(), DT)
	hdr = d.hdr.copy()
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 0
	hdr['CRVAL2'] = d.start()
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = 'aperiodicity'
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	gpkimgclass.gpk_img(hdr, ss).write(out)



#!/usr/bin/env python

import Num
import math
import hilbert_xform


def smooth(ph, dt_in, dt_out, extra=0.0, wt=None):
	"""Smooths a data set, simultaneously resampling to
	a lower sampling rate.
	This becomes inefficient if extra>>dt_out, but is otherwise
	quite fast.
	Extra has the same units as dt_in and dt_out.
	"""
	assert dt_in <= dt_out
	dt_ratio = dt_in/dt_out
	no = int(math.floor(len(ph)*dt_ratio))
	# print 'no=', dt_in, len(ph), dt_out, dt_in*len(ph)/dt_out
	dti = dt_in
	# In vari, we accumulate the width of the equivalent
	# convolution kernel that has been applied so far.
	# Specifically, it is the second moment of the kernel.
	if wt is not None:
		ph = ph * wt
		weight = wt
	else:
		weight = 1
	kwi = 0.0
	t0 = 0.0
	while dti < 0.33*dt_out:
		# print "ph=", ph
		n = ph.shape[0] - ph.shape[0]%2
		ph = ph[0:n-1:2] + ph[1:n:2]
		if wt is not None:
			weight = weight[0:n-1:2] + weight[1:n:2]
		else:
			weight *= 2
		t0 += 0.5*dti
		kwi += (dti/2.0)**2
		dti *= 2.0
	# kwo is the desired width of the convolution kernel
	# between th input and output data streams.
	# It is just the second moment of a rectangular boxcar average.
	kwo = extra**2 + dt_out**2/12.0 - dt_in**2/12.0
	# kwx is the extra smoothing kernel that is yet to be applied.
	kwx = kwo - kwi
	assert kwx >= 0.0

	# print 'ph=', ph
	# Now, we set up a Gaussian of the appropriate width
	# to do the final convolution.
	if kwx > 0.0:
		N = int(math.ceil(2.3*math.sqrt(kwx)/dti))
		i = Num.arrayrange(2*N+1) - N
		k0 = Num.exp( (-1/(2*kwx)) * (i*dti)**2 )
		assert k0.shape[0] < ph.shape[0]
		kernel = k0/Num.sum(k0)
		ph = Num.convolve(ph, kernel, 1)
		if wt is not None:
			weight = Num.convolve(weight, kernel, 1)
	Num.divide(ph, weight, ph)
	# print 'phs=', ph

	# Then grab the necessary samples.
	q = Num.arange(no)*(dt_out/dti)
	qi = Num.around(q).astype(Num.Int)
	assert qi[-1] < ph.shape[0]
	rv = Num.take(ph, qi)
	assert rv.shape[0] == no
	return (rv, t0)


def local_power(d, dt_in, dt_out, extra=0.0):
	cutoff = math.hypot(dt_out, extra)/dt_in
	ph = Num.absolute(hilbert_xform.hilbert(d, cutoff))
	return smooth(ph, dt_in, dt_out, extra)


def test1():
	x = Num.zeros((50,), Num.Float)
	x[10:] = 1.0
	k = [0.1, 0.2, 0.4, 0.2, 0.1]
	xk = Num.convolve(x, k, 1)
	xs0, ts0 = smooth(x, 1.0, 1.0, 0.01)
	xs1, ts1 = smooth(x, 1.0, 1.0, 1.0)
	xs2, ts2 = smooth(x, 1.0, 1.0, 2.0)
	for i in range(20):
		print i, x[i], xk[i], xs0[i], xs1[i], xs2[i]
		assert Num.alltrue( Num.equal(Num.less(x, 0.5),
						Num.less(xk, 0.5)
						)
					)
		assert Num.alltrue( Num.equal(Num.less(x, 0.5),
						Num.less(xs0, 0.5)
						)
					)
		assert Num.alltrue( Num.equal(Num.less(x, 0.5),
						Num.less(xs1, 0.5)
						)
					)
		assert Num.alltrue( Num.equal(Num.less(x, 0.5),
						Num.less(xs2, 0.5)
						)
					)


def test2():
	x = Num.zeros((500,), Num.Float)
	x[100:] = 1.0
	xs0, ts0 = smooth(x, 0.1, 1.0, 0.01)
	xs1, ts1 = smooth(x, 0.1, 1.0, 1.0)
	xs2, ts2 = smooth(x, 0.1, 1.0, 2.0)
	for i in range(200):
		if i%10 == 5:
			print i, x[i], xs0[i/10], xs1[i/10], xs2[i/10]
		else:
			print i, x[i]

def test3():
	x = Num.zeros((200,), Num.Float)
	x[100] = 1.0
	xs0, ts0 = smooth(x, 0.1, 0.1, 0)
	xs1, ts1 = smooth(x, 0.1, 0.3, 0)
	xs2, ts2 = smooth(x, 0.1, 0.9, 0)
	xs3, ts3 = smooth(x, 0.1, 1.5, 0)
	for i in range(197):
		o = [ '%d'%i ]
		o.append( '%g' % x[i])
		o.append( '%g' % xs0[i])
		if i%3 == 1:
			o.append('%g' % xs1[i//3])
		else:
			o.append('')

		if i%9 == 5:
			o.append('%g' % xs2[i//9])
		else:
			o.append('')

		if i%15 == 7:
			o.append('%g' % xs3[i//15])
		else:
			o.append('')

		print '\t'.join(o)


if __name__ == '__main__':
	test1()
	test2()
	test3()



#!/usr/bin/env python

"""Duration estimator for speech.

Usage: s_slope [flags]
Flags: -f FFF opens file FFF for reading.
	-c XXX  selects column XXX (either a data name or an integer).
	-o FFF  sets the output file.
	-dt #.##   Sets the interval for calculating the output.

It takes a local spectrum,
bins it onto the Bark scale,
converts to perceptual loudness (via **E).
Then, it computes a measure of how far you can go
from each point before the spectrum changes too much.
"""

import Num
import math as M
import bark_scale
import die
import gpkmisc

import sys
import os
sys.path.insert(0, os.environ['OXIVOICE'])
import voice_misc



CBmax = bark_scale.f_to_bark(5500.0)
CBmin = bark_scale.f_to_bark(100.0)
BBSZ = 0.5
SPECWINDOW = 0.0546
NORMWINDOW = 0.197
E = 0.333
C = 91.2
RTYPE = 'dur'
LFAC = 1.03
EXP1 = 2
EXP2 = 1

def NEee(x):
	# return Num.log(x)
	return x**E

def iNEee(x):
	return x**(1.0/E)


class bark_bin:
	__doc__ = """This class will bin a power spectrum in uniform frequency bins
		into bins of uniform width on a Bark scale."""

	def __init__(self, df, n2, bbsz, CBmin, CBmax):
		"""Build some arrays so we can bin data rapidly."""
		ctrs = CBmin + bbsz*Num.arrayrange(int(M.ceil((CBmax-CBmin)/bbsz)))
		lbound = bark_scale.bark_to_f(ctrs-0.5)
		assert df > 0.0
		hbound = bark_scale.bark_to_f(ctrs+0.5)
		f = Num.arrayrange(n2) * df
		llb = 0
		lhb = 0
		nf = f.shape[0]
		llba = []
		lhba = []
		correction = []
		for i in range(len(ctrs)):
			llb = int(round(lbound[i]/df))
			lhb = int(round(hbound[i]/df))
			assert lhb > llb
			if lhb >= nf:
				break
			llba.append(llb)
			lhba.append(lhb)
			assert f[lhb] > f[llb]
			correction.append((f[lhb]-f[llb])/(hbound[i]-lbound[i]))
			# This corrects for the fact that the width of each bin is not quite
			# one bark, because the input frequency data is sampled.
		self.llb = Num.array(llba, copy=True)	# Lower bound for a bin
		self.lhb = Num.array(lhba, copy=True)	# Upper bound for a bin.
		self.ctrs = ctrs[:self.llb.shape[0]]
		self.correction = Num.array(correction, copy=True)
		# print 'Bllb', self.llb
		# print 'Blhb', self.lhb
		# print 'Bcor', self.correction
		# print 'Bctr', self.ctrs
		assert Num.alltrue(Num.greater(self.correction, 0.5)*Num.less(self.correction, 2.0))
		assert self.llb.shape == self.lhb.shape
		assert self.llb.shape == self.correction.shape


	def bin(self, pwr):
		"""Do the actual binning.
		Return power in bark scale bins and bin centers.
		Bin centers are in units of barks."""

		# print 'bpwr', pwr
		pwrsum = Num.add.accumulate(pwr)
		# print 'bps', pwrsum
		o = (Num.take(pwrsum, self.lhb) - Num.take(pwrsum, self.llb))/self.correction
		# print 'bo', o
		assert Num.alltrue( Num.greater_equal(o, 0.0) )
		assert o.shape == self.ctrs.shape
		# return (o**E, self.ctrs)
		return (o, self.ctrs)


_bbc = {}
def bark_bin_cached(df, n2, bbsz, CBmin, CBmax):
	key = (df, n2, bbsz, CBmin, CBmax)
	if key not in _bbc:
		if len(_bbc) > 20:
			_bbc.popitem()
		_bbc[key] = bark_bin(df, n2, bbsz, CBmin, CBmax)
	return _bbc[key]



def _log_it(name, *d):
	fd = open(name, "w")
	n = d[0].shape
	for tmp in d:
		assert tmp.shape == n
	for i in range(n[0]):
		for tmp in d:
			fd.write("%g " % tmp[i])
		fd.write('\n')
	fd.close()



def one_spec_guts(d, extra):
	"""Approximate spectral slope of the sound in data array d.
	Extra contains misc. parameters."""

	n = d.shape[0]
	die.note("n", n)
	assert d.shape == (n,)
	df = 1/(n*extra['dt'])
	# print "# n=", n, "df=", df
	# print 'd=', d
	# print 'misc_win', voice_misc.window(n, 0)
	ss = Num.FFT.fft( d*voice_misc.window(n, 0) )
	ss[0] = 0	# Remove DC offset.
	pwr = Num.absolute(ss)**2/n
	bbin = bark_bin_cached(df, n/2, BBSZ, CBmin, CBmax)
	barkpwr, fbark = bbin.bin(pwr[:n/2])
	# print 'pwr=', Num.sum(pwr), Num.sum(barkpwr)

	return NEee(barkpwr)


def one_spec_norm(d):
	assert len(d.shape) == 1
	w = voice_misc.window(d.shape[0], 0)
	return M.sqrt(Num.sum((w*d)**2))	# Num.sum(w**2)==1


def one_spec(t, data, extra):
	"""Measure the spectrum near time t*DT.
	It returns a normalized spectrum and the normalization factor.
	"""
	# print 't=', t, 'DT=', extra['DT'], 'dt=', extra['dt']
	if extra.has_key('column'):
		extra['logname'] = "%st%d" % (extra['column'], t)
	tdt = t*extra['DT']
	s, e = voice_misc.start_end(tdt, extra['dt'],
					extra['window'], data.shape[0])
	ns, ne = voice_misc.start_end(tdt, extra['dt'],
					int(round(NORMWINDOW/dt)), data.shape[0])
	# print 's, e=', s, e, ns, ne, norm
	n1 = one_spec_norm(data[ns:ne])
	d = data[s:e]/n1
	return (one_spec_guts(d, extra), n1)



def spec(data, extra):
	"""Measure the time-series of the spectrum of a data array data.
	DT is the sampling interval for the resulting time series.
	It returns an array (s,norm) where s is the normalized spectra
	and norm is the time series of the normalized power.
	s[t,f] is the normalized power density at time t and frequency f.
	norm[t] is proportional to the power near time t."""

	ns = int(round(extra['dt']*data.shape[0]/extra['DT']))
	so = None
	no = Num.zeros((ns,), Num.Float)
	for t in range(ns):
		# t is measured relative to the beginning of the data.
		s,n = one_spec(t, data, extra)
		if so is None:
			so = Num.zeros((ns,s.shape[0]), Num.Float)
		so[t,:] = s
		no[t] = n
	return (so, no)



def calc_prelims(dt, Dt):
	extra = {'dt': dt, 'DT': Dt}
	extra['window'] = voice_misc.near_win_size(near=SPECWINDOW/dt, tol=0.1*SPECWINDOW/dt)
	return extra


def pdur_guts(s, t, dir, Dt):
	"""t is an integer; an index into the data.
	S is the normalized perceptual spectrum."""
	n = len(s)
	sumdiff = 0.0
	len_sum = 0.5	# The integral for the i==t case is trivial.
	ctr_sum = 0.125*dir	# The integral for i==t is trivial.
	i = t + dir
	while i>=0 and i<n and sumdiff<8:
		delta_diff = Dt * C * Num.sum( Num.absolute(s[i]-s[t])**EXP1 )**EXP2

		# We assume that the spectrum is s[i] in the
		# region from i-0.5 to i+0.5, and compute the
		# integral( M.exp(-integral((s(t0)-s(t''))**2 from 0 to t', dt''))
		# 	from 0 to i+0.5, dt')
		# The constancy of s over the interval (i-0.5,i+0.5)
		# makes the inner integral piecewise linear.
		# Sumdiff holds integral((s(t0)-s(t''))**2 from 0 to i-0.5)
		if delta_diff <= 0:
			# assume delta_diff = 0
			lsd = M.exp(-sumdiff)
			len_sum += lsd
			# integral( (t''-t0)*M.exp(-integral((s(t0)-s(t''))**2 from 0 to t', dt''))
			# 	from 0 to i+0.5, dt')
			slopeint = M.exp(-sumdiff) * 0.5
			ctr_sum += (i-t-0.5*dir)*lsd + dir * slopeint
			i += dir
		else:
			lsd = M.exp(-sumdiff) * (1.0-M.exp(-delta_diff))/delta_diff
			len_sum += lsd
			# integral( (t''-t0)*M.exp(-integral((s(t0)-s(t''))**2 from 0 to t', dt''))
			# 	from 0 to i+0.5, dt')
			slopeint = M.exp(-sumdiff)	\
					* (1.0-M.exp(-delta_diff)*(1+delta_diff))/delta_diff**2
			ctr_sum += (i-t-0.5*dir)*lsd + dir * slopeint
			i += dir
			sumdiff += delta_diff
		# Now, sumdiff holds integral((s(t0)-s(t''))**2 from 0 to i+0.5)
	# print 't=', t, sum
	return (len_sum*Dt, ctr_sum*Dt)


def pdur(data, extra, out):
	extra.update( calc_prelims(dt, Dt=DT))
	# print 's.n=', data.shape
	data = data - gpkmisc.N_median(data)
	data = data / M.sqrt(Num.sum( data**2 )/data.shape[0])
	sp,np = spec(data, extra)
	print 'sps=', sp.shape
	AvNormPwr = Num.sum(Num.ravel(iNEee(sp)))/sp.shape[0]
	print '# AvNormPwr=', AvNormPwr
	assert 0.03 <= AvNormPwr <= 10.0

	ns = sp.shape[0]
	o = Num.zeros((ns,), Num.Float)
	for i in range(ns):
		plp, pcp = pdur_guts(sp, i, 1, extra['DT'])
		plm, pcm = pdur_guts(sp, i, -1, extra['DT'])
		if out=='dur':
			o[i] = LFAC * (plp + plm)
		elif out=='center':
			o[i] = extra['DT']*(pcp + pcm)/(plp+plm)
		else:
			die.die('Whoops')
	return o
	


if __name__ == '__main__':
	import gpkimgclass
	try:
		import psyco
		psyco.full()
	except ImportError:
		pass

	DT = 0.01	# Seconds.  Default output sampling interval.
	arglist = sys.argv[1:]
	column = None
	signalfile = None
	outfile = None
	extrahdr = {}
	while len(arglist)>0 and arglist[0][0]=='-':
		arg = arglist.pop(0)
		if arg == '-c':
			assert column is None
			tmp = arglist.pop(0)
			try:
				column = int( tmp )
			except ValueError:
				column = tmp
		elif arg == '-f':
			signalfile = arglist.pop(0)
			die.note("signalfile", signalfile)
		elif arg == '-normwin':
			NORMWINDOW = float(arglist.pop(0))
			assert NORMWINDOW > 0
		elif arg == '-lfac':
			LFAC = M.exp(float(arglist.pop(0)))
		elif arg == '-specwin':
			SPECWINDOW = float(arglist.pop(0))
			assert SPECWINDOW > 0
		elif arg == '-exp1':
			EXP1 = float(arglist.pop(0))
			assert EXP1 > 0
		elif arg == '-exp2':
			EXP2 = float(arglist.pop(0))
			assert EXP2 > 0
		elif arg == '-C':
			C = float(arglist.pop(0))
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-dur':
			RTYPE = 'dur'
		elif arg == '-ctr':
			RTYPE = 'center'
		elif arg == '-fmax':
			CBmax = bark_scale.f_to_bark(float(arglist.pop(0)))
		elif arg == '-fmin':
			CBmin = bark_scale.f_to_bark(float(arglist.pop(0)))
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.info("Unrecognized flag: %s" % arg)
			print __doc__
			die.exit(1)
	if len(arglist) > 0:
		rootname = arglist.pop(0)
		if signalfile == None:
			signalfile = rootname + '.d'
		if outfile == None:
			outfile = rootname + '.pdur.dat'
	if outfile is None:
		outfile = 'pdur.dat'
	if signalfile is None:
		die.info("No signal file specified.")
		print __doc__
		die.exit(1)
	signal = gpkimgclass.read(signalfile)
	nch = signal.d.shape[1]
	try:
		signal.column(column)
	except KeyError:
		die.info("File has %d columns" % signal.n[1])
		die.die("Bad column: %s" % str(column))

	dt = signal.dt()
	extra = {}

	print "#column=", column
	try:
		o = pdur(signal.column(column), extra, RTYPE)
	except:
		die.warn("Exception in pdur calculation")
		raise

	hdr = signal.hdr.copy()
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 0
	hdr['CRVAL2'] = signal.start()
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = 'pseudoduration'
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	oo = gpkimgclass.gpk_img(hdr, o)
	oo.write(outfile)





#!/usr/bin/env python

"""RMS or peak amplitude time series."""


import Num
import math
import die

import sys
import os
sys.path.insert(0, os.environ['OXIVOICE'])
import voice_misc
import power
# import is_voiced



DR = 'rms'
DTSMOOTH = 0.015
DT = 0.01	# Seconds.  Default output sampling interval.
HPF = 40.0	# Hz
PKWDT = 0.020	# Full-width of window for computing peak amplitude.




def one_pk_pk(t, data, dt, DT):
	"""Measure the loudness near time t."""
	s, e = voice_misc.start_end(t*DT, dt, int(round(PKWDT/dt)), data.shape[0])
	d = data[s:e]
	return d[ Num.argmax( d ) ] - d[ Num.argmin(d) ]


def compute_intensity(data, dt, DT, DR, HPF, DTSmooth=DTSMOOTH):
	"""Measure the time-series of RMS of a data array data.
	DT is the sampling interval for the resulting RMS time series."""
	if HPF is not None:
		data = voice_misc.hipass_sym_butterworth(data, HPF*dt)
	if DR == 'rms':
		pwr, t0 = power.smooth(data**2, dt, DT, DTSmooth)
		rms = Num.sqrt(pwr)
	elif DR == 'avd':
		rms,t0 = power.smooth(Num.absolute(data), dt, DT, DTSmooth)
	elif DR == 'pkpk':
		ns = int(math.floor(data.shape[0]*(dt/DT)))
		# print 'ns=', dt, data.shape[0], DT, dt*data.shape[0]/DT
		pkpk = Num.zeros((ns,), Num.Float)
		for t in range(ns):
			pkpk[t] = one_pk_pk(t, data, dt, DT)
		rms,t0 = power.smooth(pkpk, DT, DT, DTSmooth)
	else:
		die('Bad DR=%s' % DR)
	return (rms,t0)



if __name__ == '__main__':
	import gpkimgclass
	arglist = sys.argv[1:]
	columns = None
	signalfile = None
	outfile = None
	extrahdr = {}
	while len(arglist)>0 and arglist[0][0]=='-':
		arg = arglist.pop(0)
		if arg == '-c':
			if columns == None:
				columns = []
			tmp = arglist.pop(0)
			try:
				col = int(tmp)
			except ValueError:
				col = tmp
			columns.append(col)
		elif arg == '-pkpk':
			DR = 'pkpk'
		elif arg == '-f':
			signalfile = arglist.pop(0)
			die.note("signalfile", signalfile)
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-dt':
			DT = float(arglist.pop(0))
			assert DT>0.0, 'Silly time step.'
		elif arg == '-hpf':
			HPF = float(arglist.pop(0))
			assert HPF > 0.0, "Silly highpass frequency"
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.die("Unrecognized flag: %s" % arg)
	if len(arglist) > 0:
		rootname = arglist.pop(0)
		if signalfile == None:
			signalfile = rootname + '.d'
		if outfile == None:
			outfile = '%s.%s.dat' % (rootname, DR)
	if outfile == None:
		outfile = '%s.dat' % DR
	signal = gpkimgclass.read(signalfile)
	nch = signal.d.shape[1]
	if columns == None:
		columns = range(nch)
	die.note("columns", str(columns))
	for c in columns:
		try:
			signal.column(c)
		except KeyError:
			die.info("File has %d columns" % signal.n[1])
			die.die("Bad column: %s" % str(c))

	tmp = []
	t00 = None
	for i in columns:
		tmpR, t0 = compute_intensity(signal.column(i), signal.dt(), DT, DR, HPF)
		tmp.append( tmpR )
		if t00 == None:
			t00 = t0
		else:
			assert abs(t00-t0)/DT < 0.001
	n = len(tmp[0])
	o = Num.zeros((n,), Num.Float)
	for t in tmp:
		Num.add(o, t, o)
	o = o / nch

	hdr = signal.hdr.copy()
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 0
	hdr['CRVAL2'] = signal.start() + t0
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = DR
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	oo = gpkimgclass.gpk_img(hdr, Num.transpose( [o]))
	oo.write(outfile)




#!/usr/bin/env python

import Num
import math
import ortho_poly
import die



_wincache = {}

def window(n, k):
	"""Make a window of length n with k nodes.
	The window is normalized so that Num.sum(w**2)==1."""

	assert n > 0
	assert k >= 0
	cachekey = (n,k)
	if cachekey in _wincache:
		return _wincache[ cachekey ]

	opn = ortho_poly.Chebyshev(n)
	# opn.x ranges over (-1, 1)
	w = (1 + Num.cos(opn.x*math.pi)) * opn.P(k)
	o =  w  / math.sqrt(Num.sum(w**2))

	if len(_wincache) > 20:
		_wincache.popitem()

	_wincache[ cachekey ] = o
	return o


def xexp(a):
	"""exp(a), except protected from underflows."""
	MIN = -100
	ok = Num.greater(a, MIN)
	return ok * Num.exp(Num.maximum(a, MIN))




def start_end(t, dt, window, len_d):
	"""Find start and end points for a window.
	t = desired window center (seconds).
	window = window width (in samples).
	dt = sampling rate (seconds).
	len_d = length of data array (samples).
	"""
	assert type(window) == type(1)
	# window = int(round(window/dt))
	s = int(round(t/dt - 0.5*window))
	if s < 0:
		s = 0
	elif s+window > len_d:
		s = len_d - window
	e = s + window
	return (s, e)


def _filter_win_size(n, cutoff_freq):
	np = n + max(10.0/cutoff_freq, n*0.01)
	nx = near_win_size(min=np+1, tol=1+n//11, real=True)
	return nx - nx%2


_a_factor_cache = {1:1, 2:2, 3:3}	# 1, 2, 3 *must* be in here.
def a_factor(n):	# Don't mess with private_start.
	"""Finds the smallest prime factor of a number."""

	try:
		return _a_factor_cache[n]
	except KeyError:
		pass
	f = _a_factor_guts(n)
	_a_factor_cache[n] = f
	return f


_primes = [2, 3]	# 2, 3 *must* be there.
def _a_factor_guts(n):
	for p in _primes:
		if p*p > n:
			return n
		if n%p == 0:
			return p
	d = _primes[-1] + 2
	while d*d <= n:
		if a_factor(d) == d:
			_primes.append( d )
			if n % d == 0:
				return d
		d += 2
	return n

_factor_cache = {}
def factor(n):	# Don't mess with private_start.
	"""Factors numbers into a list of prime factors,
	in increasing order.
	"""

	try:
		return _factor_cache[n]
	except KeyError:
		pass
	f = a_factor(n)
	if f == n:
		tmp = [n]
	else:
		tmp = [f] + factor(n//f)
	_factor_cache[n] = tmp[:]
	return tmp

assert factor(11) == [11]
assert factor(16) == [2, 2, 2, 2]
assert factor(100) == [2, 2, 5, 5]
assert factor(2) == [2]
assert factor(97) == [97]
assert factor(55) == [5, 11]

_fftworkcache = {}
def _fftwork(n):
	"""This estimates the work required to compute a FFT of size N,
	using the Cooley-Tukey algorithm.
	See http://en.wikipedia.org/wiki/Cooley-Tukey_FFT_algorithm
	"""
	try:
		return _fftworkcache[n]
	except KeyError:
		pass
	f1 = a_factor(n)
	if f1 == n:
		tmp = n**2
	else:
		nof1 = n/f1
		tmp = f1*_fftwork(nof1) + n + nof1*_fftwork(f1)
	_fftworkcache[n] = tmp
	return tmp


def _fftwork_real(n):
	"""This is the work for a forward and reverse transform of
	real data back to real data, using Num.FFT.rfft and Num.FFT.irfft
	"""
	return _fftwork(n) + _fftwork(n-n%2)


def near_win_size(near=None, tol=None, min=None, max=None, real=False):
	"""Find a window size near the specified size where
	the FFT algorithm is fastest."""

	if near is not None and tol is not None:
		nmax = int(round(near + tol))
		nmin = int(round(near - tol))
	elif min is not None and max is not None:
		nmax = int(round(max))
		nmin = int(round(min))
	elif min is not None and tol is not None:
		nmin = int(round(min))
		nmax = int(round(min + tol))
	elif max is not None and tol is not None:
		nmax = int(round(max))
		nmin = int(round(max - tol))
	else:
		raise ValueError, "Needs near&tol, min&max, min&tol, or max&tol."

	if real:
		fftwork = _fftwork_real
	else:
		fftwork = _fftwork

	besti = nmin
	bestwork = fftwork(besti)
	Since_lim = 100
	since = 0
	for i in range(nmin, nmax+1):
		t = fftwork(i)
		if t < bestwork:
			bestwork = t
			besti = i
			since = 0
		else:
			# This branch is just to keep near_win_size
			# from trying too hard.   Otherwise, it's quite
			# possible to spend more time finding the optimal
			# size than one spends taking the actual FFT.
			since += 1
			if since > Since_lim:
				break
	return besti


def near_win_size_real(near=None, tol=None, min=None, max=None):
	return near_win_size(near, tol, min, max, real=True)


def test_win_size():
	assert near_win_size(min=4, max=7) == 4
	assert near_win_size(min=5, max=7) != 7
	assert near_win_size(near=15, tol=2) != 17
	assert near_win_size(near=14, tol=1) != 13
	assert near_win_size(near=97, tol=1) != 97
	assert near_win_size(near=197, tol=1) != 197
	assert near_win_size(near=23, tol=1) != 23


def test_print_win():
	N = 100
	w0 = window(N, 0)
	w1 = window(N, 1)
	w2 = window(N, 2)
	w3 = window(N, 3)
	w4 = window(N, 4)
	w5 = window(N, 5)
	for i in range(N):
		print i, w0[i], w1[i], w2[i], w3[i], w4[i], w5[i]



def hipass_first_order(d, cutoff_freq):
	"""A first-order high-pass filter.
	Cutoff freq is measured in cycles per point."""

	assert cutoff_freq >= 0.0
	assert len(d.shape) == 1
	n = d.shape[0]
	nx = _filter_win_size(n, cutoff_freq)
	dx = Num.FFT.rfft(d, nx)
	f = real_fft_freq(dx.shape[0])
	fr = 1j* (f / cutoff_freq)
	ff = fr / (1 + fr)
	ff[0] = ff[0].real
	ff[-1] = ff[-1].real
	return Num.FFT.irfft(dx*ff, nx)[:n]


def lopass_first_order(d, cutoff_freq):
	"""A first-order low-pass filter.
	Cutoff freq is measured in cycles per point."""

	assert cutoff_freq >= 0.0
	assert len(d.shape) == 1
	n = d.shape[0]
	nx = _filter_win_size(n, cutoff_freq)
	dx = Num.FFT.rfft(d, nx)
	f = real_fft_freq(dx.shape[0])
	fr = 1j* (f / cutoff_freq)
	ff = 1 / (1 + fr)
	ff[0] = ff[0].real
	ff[-1] = ff[-1].real
	return Num.FFT.irfft(dx*ff, nx)[:n]


def bandpass_first_order(d, cutoff_hp, cutoff_lp):
	"""A combinatio of a first-order high-pass filter
	and a low-pass filter.
	Cutoff freq is measured in cycles per point."""

	assert cutoff_lp >= 0.0
	assert cutoff_hp >= 0.0
	assert len(d.shape) == 1
	n = d.shape[0]
	nx = _filter_win_size(n, 1.0/math.hypot(1.0/cutoff_lp, 1.0/cutoff_hp))

	dx = Num.FFT.rfft(d, nx)
	f = real_fft_freq(dx.shape[0])

	frlp = 1j* (f / cutoff_lp)
	frhp = 1j* (f / cutoff_hp)
	ff = frhp / ((1 + frhp)*(1+frlp))
	ff[0] = ff[0].real
	ff[-1] = ff[-1].real
	return Num.FFT.irfft(dx*ff, nx)[:n]


def test_hpf():
	q = Num.cos(Num.arrayrange(10000)*0.003*2*math.pi)
	sq = Num.sum(q*q)
	q1 = hipass_first_order(q, 0.0003)
	sq1 = Num.sum(q1*q1)
	assert 0.98*sq < sq1 < 0.995*sq
	q4 = hipass_first_order(q, 0.003)
	sq4 = Num.sum(q4*q4)
	assert 0.4*sq < sq4 < 0.6*sq
	q5 = hipass_first_order(q, 0.03)
	sq5 = Num.sum(q5*q5)
	# print sq5/sq
	assert  0.003 < sq5 < 0.03*sq




def fft_freq(n, d=1.0):
	"""Returns an array of indices whose frequencies are between f0 and f1
	after you've called fft().
	This will produce negative values where appropriate.
	"""
	if hasattr(Num.FFT, 'fftfreq'):
		return Num.FFT.fftfreq(n, d)
	fidx = Num.arrayrange(n)
	fneg = fidx - n
	useneg = Num.less(-fneg, fidx)
	df = 0.5/(n*d)
	return df * Num.where(useneg, fneg, fidx)


def fft_indices(f0, f1, n, d=1.0):
	"""Returns an array of indices whose frequencies are between f0 and f1
	after you've called real_fft().
	"""
	raise RuntimeError, "Sorry! Not implemented"


def real_fft_freq(n, d=1.0):
	"""Returns the frequencies associated with an index, after you've
	called real_fft().
	"""
	fidx = Num.arrayrange(n)
	df = 0.5/(n*d)
	return df * fidx


def real_fft_indices(f0, f1, n, d=1.0):
	"""Returns the indices whose frequencies are between f0 and f1
	after you've called real_fft().
	"""
	df = 0.5/(n*d)
	f0 = max(0.0, f0)
	imin = int(math.ceil(f0/df))
	imax = int(math.floor(f1/df))
	return (imin, imax+1)



def dataprep_flat_real(data, dt, edgewidth, pad=None):
	return dataprep_flat_generic(data, dt, edgewidth, pad, True)

def dataprep_flat(data, dt, edgewidth, pad=None):
	return dataprep_flat_generic(data, dt, edgewidth, pad, False)


def dataprep_flat_generic(data, dt, edgewidth, pad, isreal):
	"""Pad the data, subtract the average, and round
	the edges of the window.
	"""
	m = data.shape[0]
	if pad is None:
		pad = 2*edgewidth + m//10
	min_win_size = m + pad
	n = near_win_size(min=min_win_size, tol=m//11 + 1, real=isreal)
	d = Num.zeros((n,), Num.Float)
	avg = Num.average(data)
	d[:m] = data - avg
	nedge = int(round(edgewidth/dt))
	assert nedge <= m
	ew = 0.5*(1.0-Num.cos((math.pi/nedge)*(Num.arrayrange(nedge)+0.5)))
	Num.multiply(d[:nedge], ew, d[:nedge])
	Num.multiply(d[m-nedge:m], ew[::-1], d[m-nedge:m])
	return d


def lowpass_sym_butterworth(d, cutoff_freq, order=4):
	"""A time-symmetric filter with a magnitude response that
	is the same as the Butterworth filter.
	Cutoff freq is measured in cycles per point."""
	assert len(d.shape) == 1
	if cutoff_freq<=0 or cutoff_freq>10.0:
		die.warn('Silly value of cutoff_freq: %g cycles per point.' % cutoff_freq)
	n = d.shape[0]
	nx = _filter_win_size(n, cutoff_freq)
	dx = Num.FFT.rfft(d, nx)
	f = real_fft_freq(dx.shape[0])
	ff = Num.sqrt(1.0/(1 + (f/cutoff_freq)**(2*order)))
	return Num.FFT.irfft(dx*ff, nx)[:n]




def lowpass_sym_Gaussian(d, cutoff_freq):
	"""A time-symmetric filter with a Gaussian impulse response.
	Cutoff freq is measured in cycles per point.
	"""
	assert len(d.shape) == 1
	if cutoff_freq<=0 or cutoff_freq>10.0:
		die.warn('Silly value of cutoff_freq: %g cycles per point.' % cutoff_freq)
	n = d.shape[0]
	nx = _filter_win_size(n, cutoff_freq)
	dx = Num.FFT.rfft(d, nx)
	f = real_fft_freq(dx.shape[0])
	q = 1.0/(2*cutoff_freq**2)
	ff = Num.exp(-f**2 * q)
	return Num.FFT.irfft(dx*ff, nx)[:n]

lowpass_sym_gaussian = lowpass_sym_Gaussian



def hipass_sym_butterworth(d, cutoff_freq, order=4):
	"""A time-symmetric filter with a magnitude response that
	is the same as the Butterworth filter.
	Cutoff freq is measured in cycles per point."""
	assert len(d.shape) == 1
	if cutoff_freq<=0 or cutoff_freq>10.0:
		die.warn('Silly value of cutoff_freq: %g cycles per point.' % cutoff_freq)
	n = d.shape[0]
	nx = _filter_win_size(n, cutoff_freq)
	dx = Num.FFT.rfft(d, nx)
	f = real_fft_freq(dx.shape[0])
	ff1 = (f/cutoff_freq)**(2*order)
	ff = Num.sqrt( ff1 / (1 + ff1) )
	return Num.FFT.irfft(dx*ff, nx)[:n]


def test_lpb():
	q = Num.cos(Num.arrayrange(10000)*0.003*2*math.pi)
	sq = Num.sum(q*q)
	q1 = lowpass_sym_butterworth(q, 0.1)
	sq1 = Num.sum(q1*q1)
	# print sq1/sq
	assert 0.99*sq < sq1 < 1.00001*sq
	q2 = lowpass_sym_butterworth(q, 0.05)
	sq2 = Num.sum(q2*q2)
	# print sq2/sq
	assert 0.99*sq < sq2 < 1.00001*sq
	q4 = lowpass_sym_butterworth(q, 0.003)
	sq4 = Num.sum(q4*q4)
	# print sq4/sq
	assert 0.4*sq < sq4 < 0.6*sq
	q5 = lowpass_sym_butterworth(q, 0.001)
	sq5 = Num.sum(q5*q5)
	# print sq5/sq
	assert  0 < sq5 < 0.2*sq
	

def test_hpb():
	q = Num.cos(Num.arrayrange(10000)*0.003*2*math.pi)
	sq = Num.sum(q*q)
	q1 = hipass_sym_butterworth(q, 0.0003)
	sq1 = Num.sum(q1*q1)
	# print sq1/sq
	assert 0.99*sq < sq1 < 1.00001*sq
	q4 = hipass_sym_butterworth(q, 0.003)
	sq4 = Num.sum(q4*q4)
	# print sq4/sq
	assert 0.4*sq < sq4 < 0.6*sq
	q5 = hipass_sym_butterworth(q, 0.03)
	sq5 = Num.sum(q5*q5)
	# print sq5/sq
	assert  0 < sq5 < 0.2*sq


def pink_noise(n):
	"""Create pink noise where the power spectral density
	goes as 1/f (except right at f=0).
	"""
	m = near_win_size_real(near=2*n+2, tol=n//3+1)

	rtmp = Num.RA.standard_normal(m)
	itmp = Num.RA.standard_normal(m)
	tmp = rtmp + 1j*itmp
	f = real_fft_freq(m)
	assert f[0] == 0.0
	tmp[0] = 0.0
	f[0] = 1.0
	Num.multiply(tmp, 1.0/Num.sqrt( f ), tmp)
	return Num.FFT.inverse_real_fft(tmp)[:n]

if __name__ == '__main__':
	test_win_size()
	test_print_win()
	test_hpf()
	test_lpb()
	test_hpb()



#!/usr/bin/env python

"""This simulates re-running feature_vec.py with the -xform or -XFORM flags.
It reads in a feature vector, then transforms it by a matrix multiplication,
and writes it back out.
"""

import math
import chunkio
import gpkimgclass
import avio
import Num
import os
import sys
sys.path.insert(0, os.environ['OXIVOICE'])
import feature_vec as FV

PLOT = False

XFORM = None


def fiddle(h):
	o = {}
	for (k, v) in h.items():
		if k.startswith('F_INFO'):
			o['OLD_%s' % k] = v
		elif k.startswith('RMS'):
			pass
		elif k.startswith('TTYPE'):
			pass
		elif k.startswith('TUNITS'):
			pass
		else:
			o[k] = v
	return o


def process(d, XFORM, verbose=0):
	fac = chunkio.datachunk(open(XFORM, 'r')).read_NumArray()
	o = Num.matrixmultiply(data.d, fac)
	hdr = fiddle( data.hdr )
	hdr['XFORM'] = XFORM
	for i in range( fac.shape[1] ):
		hdr['TTYPE%d' % (i+1)] = 'Mixture%d' % i
		d = ','.join(['%g' % q for q in fac[:,i] ])
		hdr['F_INFO%d' % (i+1)] = d
		hdr['RMS%d' % (i+1)] = math.sqrt(Num.average(o[:,i]**2)) 
	gpkimgclass.gpk_img(hdr, o).write(ofile)



if __name__ == '__main__':
	ofile = '-'
	arglist = sys.argv[1:]
	verbose = 0
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-o':
			ofile = arglist.pop(0)
		elif arg == '-v':
			verbose += 1
		elif arg == '-xform':
			fxd = os.environ['OXIVOICE']
			XFORM = '%s/xform.chunk' % fxd
		elif arg == '-Xform':
			XFORM = arglist.pop(0)
		else:
			die.die('Unrecognized flag: %s' % arg)
	assert XFORM, "Nothing to do!"
	fname = arglist[0]
	data = gpkimgclass.read(fname)
	process(data, XFORM, verbose=verbose)



#!/usr/bin/env python


import erb_scale
import Num
import die
import math
import avio
import gpkmisc
import sets

import os
import sys
sys.path.insert(0, os.environ['OXIVOICE'])
import percep_spec
import voice_misc
# import irregularity
# sys.path.insert(0, '%s/tick1/bin' % os.environ['MRImodel'])

if 'PYLABDISP' in os.environ:
	if os.environ['PYLABDISP'] == 'pylab':
		import pylab
	else:
		import g_pylab as pylab
	PLOT = True
else:
	PLOT = False

if False:
	def edgepair_win(n):
		e = Num.zeros((2*n,), Num.Float)
		e[:n] = -1.0
		e[n:] = 1.0
		s = Num.ones((2*n), Num.Float)
		return (e, s)
else:
	def edgepair_win(n):
		tmp0 = voice_misc.window(2*n, 0)
		tmp1 = voice_misc.window(2*n, 1)
		norm = gpkmisc.N_maximum(tmp0)
		return (tmp1/norm, tmp0/norm)


if False:
	def win(n):
		tmp = Num.ones((n,), Num.Float)
		return tmp
else:
	def win(n):
		tmp = voice_misc.window(n, 0)
		return tmp/gpkmisc.N_maximum(tmp)


def burst_win(n, i):
	tmp = Num.zeros((n,), Num.Float)
	tmp[i] = 1.0
	neg = win(n)
	nneg = neg/Num.sum(neg)
	return (tmp - nneg, nneg)


def entropy(w):
	assert Num.alltrue( Num.greater(w, 0.0) )
	p = w/Num.sum(w)
	return -Num.sum(p*Num.log(p))



# SpecExp = 0.0
Nsv = 0.75
DT = 0.005
DB = math.sqrt(0.5)
LF = 1.5
ELF = 1.5
# Nsv and LF updated from the 5 April 2007 optimization
# in .../m/ASR/OPT

def feature_vec(data, dt, DT=DT,
		LF=1.0, Nsv=Nsv,
		ELF=1.0,
		do_voicing=True, do_dissonance=True):
	FORMANT_LOW = erb_scale.f_to_erb(60.0)
	FORMANT_HIGH = erb_scale.f_to_erb(6000.0)
	assert DT > 0.0 and float(DT)>0.0
	assert LF > 0.0 and float(LF)>0.0
	bmin = erb_scale.f_to_erb(80.0)
	bmax = erb_scale.f_to_erb(7000.0)
	all_ectrs, all_ps, t0 = percep_spec.perceptual_spec(data, dt, DT,
							bmin, bmax, DB,
							do_mod=do_voicing,
							do_dissonance=do_dissonance
							)

	band_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']=='band']
	neural = all_ps.take(band_indices, axis=0)
	ectrs = [ec for ec in all_ectrs if ec['type']=='band']
	nband_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']!='band']
	nneural = all_ps.take(nband_indices, axis=0)
	nectrs = [ec for ec in all_ectrs if ec['type']!='band']

	assert nneural.shape[1]==neural.shape[1]
	assert neural.shape[1]==all_ps.shape[1]
	assert neural.shape[0]+nneural.shape[0] == all_ps.shape[0]

	neural_now = Num.average(neural, axis=0)	# Average over frequency.
	assert neural_now.shape[0] == neural.shape[1]
	neural_avg = Num.sum(neural_now**2)/Num.sum(neural_now)	# Average over time.
		# neural_avg is a scalar, grand average.
	Num.divide(neural, neural_avg, neural)
		# Now, we've normalized by an over-all average loudness.
	Num.divide(neural_now, neural_avg, neural_now)
		# Now, we've normalized by an over-all average loudness.

	assert nneural.shape[0] < nneural.shape[1]
	assert len(nectrs) == nneural.shape[0]
	for (i,e) in enumerate(nectrs):
		if e['type'] == 'haspitch':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])
		if e['type'] == 'dissonance':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])

	# print '# neural_avg=', neural_avg
	o = []
	descr = []
	wset = sets.Set([0])
	for vl in [0.01]:
		# print 'vl=', vl, type(vl), 'LF=', LF, type(LF), 'DT=', DT, type(DT)
		w = int(round(vl*LF/DT))
		if not w in wset:
			tmpo, tmpd = vowel(w, ectrs, neural, neural_now, Nsv,
						formant_low=FORMANT_LOW,
						formant_high=FORMANT_HIGH
						)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)

	wset = sets.Set([0])
	for fel in [0.06]:
		w = int(round(fel*ELF/DT))
		if not w in wset:
			tmpo, tmpd = fricative_edge(ectrs, neural, neural_now, Nsv,
						do_abs=False, width=w
						)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)

	wset = sets.Set([0])
	for sel in [0.01]:
		w = int(round(sel*LF/DT))
		if not w in wset:
			tmpo, tmpd = spectral_entropy(w, ectrs, neural, neural_now, Nsv)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
		assert len(descr)==len(o), "Descriptor mismatch"


	if do_voicing:
		wset = sets.Set([0])
		for hpl in [0.06]:
			w = int(round(hpl*LF/DT))
			if not w in wset:
				tmpo, tmpd = haspitch(w, nectrs, nneural, neural_now, Nsv)
				o.extend(tmpo)
				descr.extend(tmpd)
				wset.add(w)
	if do_dissonance:
		wset = sets.Set([0])
		for dsl in [0.06]:
			w = int(round(dsl*LF/DT))
			if not w in wset:
				tmpo, tmpd = dissonance(w, nectrs, nneural, neural_now, Nsv)
				o.extend(tmpo)
				descr.extend(tmpd)
				wset.add(w)

	assert len(descr)==len(o), "Descriptor mismatch"
	N = neural[0].shape[0]
	for (i, (tmp, dsc)) in enumerate(zip(o, descr)):
		assert tmp.shape == (N,), "Wrong size: %d, %s = %d vs. %d" % (i, str(dsc), tmp.shape[0], N)

	return (o, descr, DT, t0)



def feature_vec20071030(data, dt, DT=DT,
		LF=1.0, Nsv=Nsv,
		ELF=1.0,
		do_voicing=True, do_dissonance=True):
	FORMANT_LOW = erb_scale.f_to_erb(60.0)
	FORMANT_HIGH = erb_scale.f_to_erb(5000.0)
	assert DT > 0.0 and float(DT)>0.0
	assert LF > 0.0 and float(LF)>0.0
	bmin = erb_scale.f_to_erb(80.0)
	bmax = erb_scale.f_to_erb(7000.0)
	all_ectrs, all_ps, t0 = percep_spec.perceptual_spec(data, dt, DT,
							bmin, bmax, DB,
							do_mod=do_voicing,
							do_dissonance=do_dissonance
							)

	band_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']=='band']
	neural = all_ps.take(band_indices, axis=0)
	ectrs = [ec for ec in all_ectrs if ec['type']=='band']
	nband_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']!='band']
	nneural = all_ps.take(nband_indices, axis=0)
	nectrs = [ec for ec in all_ectrs if ec['type']!='band']

	assert nneural.shape[1]==neural.shape[1]
	assert neural.shape[1]==all_ps.shape[1]
	assert neural.shape[0]+nneural.shape[0] == all_ps.shape[0]

	neural_now = Num.average(neural, axis=0)	# Average over frequency.
	assert neural_now.shape[0] == neural.shape[1]
	neural_avg = Num.sum(neural_now**2)/Num.sum(neural_now)	# Average over time.
		# neural_avg is a scalar, grand average.
	Num.divide(neural, neural_avg, neural)
		# Now, we've normalized by an over-all average loudness.
	Num.divide(neural_now, neural_avg, neural_now)
		# Now, we've normalized by an over-all average loudness.

	assert nneural.shape[0] < nneural.shape[1]
	assert len(nectrs) == nneural.shape[0]
	for (i,e) in enumerate(nectrs):
		if e['type'] == 'haspitch':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])
		if e['type'] == 'dissonance':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])

	# print '# neural_avg=', neural_avg
	o = []
	descr = []
	wset = sets.Set([0])
	for vl in [0.04]:
		# print 'vl=', vl, type(vl), 'LF=', LF, type(LF), 'DT=', DT, type(DT)
		w = int(round(vl*LF/DT))
		if not w in wset:
			tmpo, tmpd = vowel(w, ectrs, neural, neural_now, Nsv,
						formant_low=FORMANT_LOW,
						formant_high=FORMANT_HIGH
						)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)

	wset = sets.Set([0])
	for fel in [0.06]:
		w = int(round(fel*ELF/DT))
		if not w in wset:
			tmpo, tmpd = fricative_edge(ectrs, neural, neural_now, Nsv,
						do_abs=False, width=w
						)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)

	wset = sets.Set([0])
	for sel in [0.04]:
		w = int(round(sel*LF/DT))
		if not w in wset:
			tmpo, tmpd = spectral_entropy(w, ectrs, neural, neural_now, Nsv)
			o.extend(tmpo)
			descr.extend(tmpd)
			wset.add(w)
		assert len(descr)==len(o), "Descriptor mismatch"


	if do_voicing:
		wset = sets.Set([0])
		for hpl in [0.05]:
			w = int(round(hpl*LF/DT))
			if not w in wset:
				tmpo, tmpd = haspitch(w, nectrs, nneural, neural_now, Nsv)
				o.extend(tmpo)
				descr.extend(tmpd)
				wset.add(w)
	if do_dissonance:
		wset = sets.Set([0])
		for dsl in [0.05]:
			w = int(round(dsl*LF/DT))
			if not w in wset:
				tmpo, tmpd = dissonance(w, nectrs, nneural, neural_now, Nsv)
				o.extend(tmpo)
				descr.extend(tmpd)
				wset.add(w)

	assert len(descr)==len(o), "Descriptor mismatch"
	N = neural[0].shape[0]
	for (i, (tmp, dsc)) in enumerate(zip(o, descr)):
		assert tmp.shape == (N,), "Wrong size: %d, %s = %d vs. %d" % (i, str(dsc), tmp.shape[0], N)

	return (o, descr, DT, t0)



def simple_feature_vec(data, dt, DT=DT,
		LF=1.0, Nsv=Nsv, ELF=None,
		do_voicing=True, do_dissonance=True):
	FORMANT_LOW = erb_scale.f_to_erb(60.0)
	FORMANT_HIGH = erb_scale.f_to_erb(6000.0)
	assert DT > 0.0 and float(DT)>0.0
	assert LF > 0.0 and float(LF)>0.0
	bmin = erb_scale.f_to_erb(80.0)
	bmax = erb_scale.f_to_erb(7000.0)
	all_ectrs, all_ps, t0 = percep_spec.perceptual_spec(data, dt, DT,
							bmin, bmax, DB,
							do_mod=do_voicing,
							do_dissonance=do_dissonance
							)

	band_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']=='band']
	# print 'band_indices=', band_indices
	neural = all_ps.take(band_indices, axis=0)
	ectrs = [ec for ec in all_ectrs if ec['type']=='band']
	# print 'ectrs=', ectrs
	nband_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']!='band']
	# print 'nband_indices=', nband_indices
	nneural = all_ps.take(nband_indices, axis=0)
	nectrs = [ec for ec in all_ectrs if ec['type']!='band']
	# print 'nectrs=', nectrs

	assert nneural.shape[1]==neural.shape[1]
	assert neural.shape[1]==all_ps.shape[1]
	assert neural.shape[0]+nneural.shape[0] == all_ps.shape[0]

	neural_now = Num.average(neural, axis=0)
	# pylab.plot(neural_now)
	# pylab.title('neural_now')
	# pylab.show()
	assert neural_now.shape[0] == neural.shape[1]
	neural_avg = Num.sum(neural_now**2)/Num.sum(neural_now)
	# print 'neural_avg=', neural_avg
	Num.divide(neural, neural_avg, neural)
	Num.divide(neural_now, neural_avg, neural_now)
	# pylab.plot(neural_now)
	# pylab.title('normed neural_now')
	# pylab.show()

	assert nneural.shape[0] < nneural.shape[1]
	assert len(nectrs) == nneural.shape[0]
	for (i,e) in enumerate(nectrs):
		if e['type'] == 'haspitch':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])
		if e['type'] == 'dissonance':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])

	# print '# neural_avg=', neural_avg
	o = []
	descr = []
	tmpo, tmpd = vowel(1, ectrs, neural, neural_now, Nsv,
						formant_low=FORMANT_LOW,
						formant_high=FORMANT_HIGH
						)
	o.extend(tmpo)
	descr.extend(tmpd)
	assert len(descr)==len(o), "Descriptor mismatch"

	tmpo, tmpd = spectral_entropy(1, ectrs, neural, neural_now, Nsv)
	o.extend(tmpo)
	descr.extend(tmpd)
	assert len(descr)==len(o), "Descriptor mismatch"


	if do_voicing:
		Hpl = 0.03
		w = int(round(Hpl*LF/DT))
		tmpo, tmpd = haspitch(w, nectrs, nneural, neural_now, Nsv)
		o.extend(tmpo)
		descr.extend(tmpd)

	if do_dissonance:
		Dsl = 0.04
		w = int(round(Dsl*LF/DT))
		tmpo, tmpd = dissonance(w, nectrs, nneural, neural_now, Nsv)
		o.extend(tmpo)
		descr.extend(tmpd)

	assert len(descr)==len(o), "Descriptor mismatch"
	N = neural[0].shape[0]
	for (i, (tmp, dsc)) in enumerate(zip(o, descr)):
		assert tmp.shape == (N,), "Wrong size: %d, %s = %d vs. %d" % (i, str(dsc), tmp.shape[0], N)

	return (o, descr, DT, t0)



def ss_feature_vec(data, dt, DT=DT,
		LF=1.0, Nsv=Nsv, ELF=None,
		do_voicing=True, do_dissonance=False):
	FORMANT_LOW = erb_scale.f_to_erb(200.0)
	FORMANT_HIGH = erb_scale.f_to_erb(4000.0)
	assert DT > 0.0 and float(DT)>0.0
	assert LF > 0.0 and float(LF)>0.0
	bmin = erb_scale.f_to_erb(100.0)
	bmax = erb_scale.f_to_erb(6000.0)
	all_ectrs, all_ps, t0 = percep_spec.perceptual_spec(data, dt, DT,
							bmin, bmax, DB,
							do_mod=True,
							do_dissonance=False
							)

	band_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']=='band']
	# print 'band_indices=', band_indices
	neural = all_ps.take(band_indices, axis=0)
	ectrs = [ec for ec in all_ectrs if ec['type']=='band']
	# print 'ectrs=', ectrs
	nband_indices = [i for (i,ec) in enumerate(all_ectrs) if ec['type']!='band']
	# print 'nband_indices=', nband_indices
	nneural = all_ps.take(nband_indices, axis=0)
	nectrs = [ec for ec in all_ectrs if ec['type']!='band']
	# print 'nectrs=', nectrs

	assert nneural.shape[1]==neural.shape[1]
	assert neural.shape[1]==all_ps.shape[1]
	assert neural.shape[0]+nneural.shape[0] == all_ps.shape[0]

	neural_now = Num.average(neural, axis=0)
	# pylab.plot(neural_now)
	# pylab.title('neural_now')
	# pylab.show()
	assert neural_now.shape[0] == neural.shape[1]
	neural_avg = Num.sum(neural_now**2)/Num.sum(neural_now)
	# print 'neural_avg=', neural_avg
	Num.divide(neural, neural_avg, neural)
	Num.divide(neural_now, neural_avg, neural_now)
	# pylab.plot(neural_now)
	# pylab.title('normed neural_now')
	# pylab.show()

	assert nneural.shape[0] < nneural.shape[1]
	assert len(nectrs) == nneural.shape[0]
	for (i,e) in enumerate(nectrs):
		if e['type'] == 'haspitch':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])
		if e['type'] == 'dissonance':
			Num.divide(nneural[i,:], neural_avg, nneural[i,:])

	# print '# neural_avg=', neural_avg
	o = []
	descr = []
	Vl = 0.06
	w = int(round(Vl*LF/DT))
	tmpo, tmpd = vowel(w, ectrs, neural, neural_now, Nsv,
						formant_low=FORMANT_LOW,
						formant_high=FORMANT_HIGH
						)
	o.extend(tmpo)
	descr.extend(tmpd)
	assert len(descr)==len(o), "Descriptor mismatch"

	if not do_voicing:
		raise RuntimeError, "Assumes voicing=True"
	if do_dissonance:
		raise RuntimeError, "Assumes dissonance=False"

	Hpl = 0.07
	w = int(round(Hpl*LF/DT))
	tmpo, tmpd = haspitch(w, nectrs, nneural, neural_now, Nsv)
	o.extend(tmpo)
	descr.extend(tmpd)

	assert len(descr)==len(o), "Descriptor mismatch"
	N = neural[0].shape[0]
	for (i, (tmp, dsc)) in enumerate(zip(o, descr)):
		assert tmp.shape == (N,), "Wrong size: %d, %s = %d vs. %d" % (i, str(dsc), tmp.shape[0], N)

	return (o, descr, DT, t0)


def isfloat(x):
	return isinstance(x, float)


def convolve(signal, kernel):
	"""Returns something the length of the signal, by zero padding."""
	if signal.shape[0] > kernel.shape[0]:
		return Num.convolve(signal, kernel, 1)
	m = 1+kernel.shape[0]-signal.shape[0]
	tmp = Num.concatenate((signal, Num.zeros((m,), Num.Float)))
	die.info('Narrow signal: padding from %d to %d' % (signal.shape[0], m))
	return Num.convolve(tmp, kernel, 1)[:signal.shape[0]]


def vowel(width, ectrs, neural, neural_now, Nsv,
		formant_low=None, formant_high=None):
	VFAC = 0.75
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# The second term within the hypot() should always be Nsv*one.

	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type']=='band' and formant_low < e['erb'] < formant_high:
			tmp = (VFAC/css)*convolve(neural[i], cs)/norm
			# pylab.plot(tmp)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['type'] = 'vowel'
			dtmp['width'] = width
			dtmp['Kentropy'] = entropy(cs)
			dtmp['Fentropy'] = 0.0
			dtmp['id'] = 'vowel:%d:%.1f' % (width, e['erb'])
			# pylab.title('vowel %s' % dtmp['id'])
			# pylab.figure()
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def spectral_entropy(width, ectrs, neural, neural_now, Nsv):
	"""This is inspired (but only loosely) by
	"Robust Entropy-based Endpoint Detection for Speech Recognition
	in Noisy Environments." by Jia-lin Shen and Jeih-weih Hung
	and Lin-shan Lee, http://www.ee.columbia.edu/~dpwe/papers/ShenHL98-endpoint.pdf
	International Conference on Spoken Language Processing, 1998.
	"""
	assert Nsv > 0.0
	SEF = 0.5
	EPS = 1e-10
	cs = win(width)
	assert Num.alltrue( Num.greater(cs, 0.0) )
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/css
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	assert Num.alltrue( Num.greater(norm, 0.0) )
	# pylab.plot(norm)
	# pylab.figure()

	ent_sum = None
	np = 0
	psum = None
	Ne = len(ectrs)
	for (i, e) in enumerate(ectrs):
		if e['type']=='band':
			assert Num.alltrue( Num.greater_equal(neural[i], 0.0) )
			ptmp = ((1.0+Nsv)/(Ne*css)) * convolve(neural[i], cs)/norm
			# pylab.plot(ptmp)
			tmp = ptmp * Num.log(ptmp+EPS)
			if ent_sum is None:
				ent_sum = Num.array(tmp, copy=True)
			else:
				Num.add(ent_sum, tmp, ent_sum)
			np += 1

	# pylab.title('spectral_entropy')
	# pylab.show()
	assert np > 0
	dtmp = {'type': 'spectral_entropy',
		'width': width,
		'Kentropy': entropy(cs),
		'Fentropy': math.log(np),
		'id': 'entropy:%d' % width
		}
	EFAC = -0.4
	es0 = EFAC*math.log(np)
	return ([-SEF*(ent_sum-es0)], [dtmp])



def haspitch(width, ectrs, neural, neural_now, Nsv):
	HFAC = {2: 0.08, 1: 0.28}
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type'].startswith('haspitch'):
			tmp = (HFAC[e['variant']]/css)*convolve(neural[i], cs)/norm
			# pylab.figure()
			# pylab.title('e=%s' % e['id'])
			# pylab.plot(neural[i])
			# pylab.plot(tmp)
			# pylab.plot(norm)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['id'] = '%s:%d' % (dtmp['id'], width)
			dtmp['Kentropy'] = entropy(cs)
			assert 'Fentropy' in dtmp
			dtmp['width'] = width
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def dissonance(width, ectrs, neural, neural_now, Nsv):
	HFAC = {1: 0.2}
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type'].startswith('dissonance'):
			tmp = (HFAC[e['variant']]/css)*convolve(neural[i], cs)/norm
			# pylab.figure()
			# pylab.title('e=%s' % e['id'])
			# pylab.plot(neural[i])
			# pylab.plot(tmp)
			# pylab.plot(norm)
			o.append(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['id'] = '%s:%d' % (dtmp['id'], width)
			dtmp['Kentropy'] = entropy(cs)
			assert 'Fentropy' in dtmp
			dtmp['width'] = width
			descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def vowel_edge(width, ectrs, neural, neural_now, Nsv, do_abs=False,
			FORMANT_LOW=None, FORMANT_HIGH=None):
	VEfac = 0.7
	ce, cs = edgepair_win(width)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (i, e) in enumerate(ectrs):
		if e['type']=='band' and FORMANT_LOW < e['erb'] < FORMANT_HIGH:
			tmp = VEfac * convolve(neural[i], ce)/norm
			# pylab.plot(tmp)
			# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
			dtmp = e.copy()
			dtmp['width'] = 2*width
			dtmp['Kentropy'] = entropy(cs)
			dtmp['Fentropy'] = 0.0
			if do_abs:
				o.append( Num.absolute(tmp) )
				dtmp['type'] = 'vowel |edge|'
				dtmp['id'] = 'vowel |edge|:%d:%.1f' % (width, e['erb'])
				descr.append( dtmp )
			else:
				o.append( tmp )
				dtmp['type'] = 'vowel edge'
				dtmp['id'] = 'vowel edge:%d:%.1f' % (width, e['erb'])
				descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def fricative(width, ectrs, neural, neural_now, Nsv):
	CSSE = 0.7
	N = neural[0].shape[0]
	cs = win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/css
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	for (rs, re, fac) in [ (70.0, 600.0, 0.6), (600.0, 2000.0, 0.6),
				(2000.0, 3000.0, 0.6), (3000.0, 6000.0, 0.6) ]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		included = []
		nq = 0
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append(e)
				nq += 1
		tmp = (fac/css**CSSE) * convolve(tsum, cs)/(norm*nq)
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
		# pylab.plot(tmp + len(o))
		o.append( tmp )
		descr.append( {'type': 'fricative', 'width': width,
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'id': 'fricative:%d:%.1f-%.1f' % (width, elow, ehigh)
				}
				)
	# pylab.show()
	return (o, descr)


def prominence(ectrs, neural, neural_now, Nsv):
	WIDTH = 20
	N = neural[0].shape[0]
	ce, cs = edgepair_win(WIDTH)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	included = []
	for (rs, re, fac) in [ (70.0, 600.0, 0.22), (600.0, 2000.0, 0.25),
				(2000.0, 3000.0, 0.32), (3000.0, 6000.0, 0.35) ]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append(e)
		tmp = (fac/css) * convolve(tsum, ce)/norm
		# pylab.plot(tmp + len(o))
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
		o.append( tmp )
		dtmp = {'type': 'prominence', 'width': WIDTH,
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'id': 'prominence:%d:%.1f-%.1f' % (WIDTH, elow, ehigh)
				}
		descr.append( dtmp )
	# pylab.show()
	return (o, descr)




def fricative_edge(ectrs, neural, neural_now, Nsv, do_abs=False, width=2):
	CSSE = 1.6
	N = neural[0].shape[0]
	ce, cs = edgepair_win(width)
	css = Num.sum(cs)
	nns = convolve(neural_now, cs)/css
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	included = []
	for (rs, re, fac) in [(70.0, 800.0, 7), (800.0, 1500.0, 7),
				(1500.0, 2200.0, 7), (2200.0, 6000.0, 7)]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		nq = 0
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append(e)
				nq += 1
		tmp = (fac/css**CSSE) * convolve(tsum, ce)/(norm*nq)
		# pylab.plot(tmp)
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])

		dtmp = {'width': width, 
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included))
				}

		if do_abs:
			o.append( Num.absolute(tmp) )
			dtmp['type'] = 'fricative |edge|'
		else:
			o.append( tmp )
			dtmp['type'] = 'fricative edge'
		dtmp['id'] = '%s:%d:%.1f-%.1f' % (dtmp['type'], width, elow, ehigh)
		descr.append( dtmp )
	# pylab.show()
	return (o, descr)


def burst(ectrs, neural, neural_now, Nsv):
	BKGWIDTH = 8
	WIDTH = 1
	N = neural[0].shape[0]
	ce,cs = burst_win(BKGWIDTH, 2)
	nns = convolve(neural_now, cs)/Num.sum(cs)
	norm = Num.hypot(nns, Nsv * Num.sum(nns**2)/Num.sum(nns))
	# pylab.plot(norm)
	# pylab.figure()
	o = []
	descr = []
	included = []
	# Burst bandwidth needs to be large enough so that the
	# burst can fit into a 10ms slot.
	for (rs, re, fac) in [ (150.0, 500.0, 1.0), (500.0, 1200.0, 1.0),
				(1200.0, 2800.0, 1.0), (2800.0, 6000.0, 1.0) ]:
		tsum = Num.zeros((N,), Num.Float)
		elow = erb_scale.f_to_erb(rs)
		ehigh = erb_scale.f_to_erb(re)
		for (i, e) in enumerate(ectrs):
			if e['type']=='band' and elow < e['erb'] < ehigh:
				Num.add(tsum, neural[i], tsum)
				included.append( e )
		tss = fac * convolve(tsum, ce)/norm
		# pylab.plot(tmp + len(o))
		# print 'RMS', math.sqrt(Num.sum(tmp**2)/tmp.shape[0])
		o.append( tss )
		descr.append( {'type': 'burst', 'width': WIDTH,
				'bkgwidth': BKGWIDTH,
				'erbs': [e['erb'] for e in included],
				'fcs' : [e['fc'] for e in included],
				'Kentropy': entropy(cs),
				'Fentropy': -math.log(len(included)),
				'id': 'burst:%d:%.1f-%.1f'
					% (WIDTH, elow, ehigh)
				}
				)
	# pylab.show()
	return (o, descr)



import gpkavg

def average(x):
	vv = []
	ww = []
	for (v,wt) in x:
		vv.append(v)
		ww.append(wt)
	return gpkavg.avg(vv, ww, 0.0)[0]

def median(x):
	vv = []
	ww = []
	for (v,wt) in x:
		vv.append(v)
		ww.append(wt)
	return gpkavg.avg(vv, ww, 0.499)[0]


def weight(x):
	sum = 0.0
	for (v,wt) in x:
		sum += wt**2
	return math.sqrt(sum)



def describe_xform(descr, fac):
	nin, nout = fac.shape
	assert len(descr) == nin
	for i in range(nout):
		norm = math.sqrt(Num.average(fac[:,i]**2))
		nf = fac[:,i]/norm
		tmp = []
		fwt = []
		wwt = []
		edginess = []
		voiciness = []
		burstiness = []
		friciness = []
		vness = []
		for j in range(nin):
			dj = descr[j]
			if dj.has_key('erb'):
				fwt.append( (dj['erb'], abs(nf[j])) )
			if dj.has_key('width'):
				wwt.append( (dj['width'], abs(nf[j])) )
			if 'edge' in dj['type']:
				edginess.append( abs(nf[j]) )
			if 'burst' in dj['type']:
				burstiness.append( abs(nf[j]) )
			if 'fricative' in dj['type']:
				friciness.append( abs(nf[j]) )
			if 'vowel' in dj['type']:
				vness.append( abs(nf[j]) )
			if 'haspitch' in dj['type']:
				voiciness.append( abs(nf[j]) )
			if abs(nf[j]) > 0.1:
				tmp.append('%+.2f*%s' % (nf[j], dj['id']) )
		print '# FV[%d]= %s' % (i, ' '.join(tmp))
		fb = {'voicing': math.sqrt(Num.sum(Num.array(voiciness)**2)),
			'edges': math.sqrt(Num.sum(Num.array(edginess)**2)),
			'burst': math.sqrt(Num.sum(Num.array(burstiness)**2)),
			'fricative': math.sqrt(Num.sum(Num.array(friciness)**2)),
			'vowel': math.sqrt(Num.sum(Num.array(vness)**2)),
			'meanfreq': average(fwt), 'medfreq': median(fwt),
			'wt_freq': weight(fwt),
			'meanwidth': average(wwt), 'medwidth': median(wwt),
			'wt_width': weight(wwt),
			'channel': i, 'description': ' '.join(tmp),
			'ltype': 'fb'
			}
		print avio.concoct(fb)



__version__ = "$Revision: 1.65 $"
__date__ = "$Date: 2007/04/05 15:34:43 $"

XFORM = None

if __name__ == '__main__':
	import gpkimgclass
	try:
		import psyco
		psyco.full()
	except ImportError:
		pass
	ofile = '-'
	arglist = sys.argv[1:]
	verbose = 0
	column = 0
	DoVoicing = True
	DoDissonance = True
	fvf = feature_vec
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-o':
			ofile = arglist.pop(0)
		elif arg == '-v':
			verbose += 1
		elif arg == '-Nsv':
			Nsv = float(arglist.pop(0))
		# elif arg == '-SpecExp':
			# SpecExp = float(arglist.pop(0))
		elif arg == '-col':
			column = arglist.pop(0)
		elif arg == '-DT':
			DT = float(arglist.pop(0))
		elif arg == '-sfv':
			fvf = simple_feature_vec
		elif arg == '-ssfv':
			fvf = ss_feature_vec
			DoDissonance = False
		elif arg == '-fv20071030':
			fvf = feature_vec20071030
		elif arg == '-DB':
			DB = float(arglist.pop(0))
		elif arg == '-LF':
			LF = float(arglist.pop(0))
		elif arg == '-ELF':
			ELF = float(arglist.pop(0))
		elif arg == '-novoicing':
			DoVoicing = False
		elif arg == '-xform':
			fxd = os.environ['OXIVOICE']
			XFORM = '%s/xform.chunk' % fxd
		elif arg == '-Xform':
			XFORM = arglist.pop(0)
		else:
			die.die('Unrecognized flag: %s' % arg)
	fname = arglist[0]
	data = gpkimgclass.read(fname)
	o, descr, DT, t0 = fvf(data.column(column), data.dt(),
					Nsv=Nsv,
					# SpecExp=SpecExp,
					LF=LF, DT=DT, ELF=ELF,
					do_voicing=DoVoicing,
					do_dissonance=DoDissonance
					)
	o = Num.transpose(o)
	hdr = {'CDELT2': DT, 'CRPIX2': 1, 'CRVAL2': data.start()+t0,
		'VERSION': __version__,
		'BITPIX': -32, 'Nsv': Nsv,
		'FNAME': fname,
		# 'SpecExp': SpecExp,
		'DataSamplingFreq': 1.0/data.dt(),
		'LF': LF, 'Nsv': Nsv, 'ELF': ELF
		}
	if XFORM is not None:
		import chunkio
		fac = chunkio.datachunk(open(XFORM, 'r')).read_NumArray()
		o = Num.matrixmultiply(o, fac)
		if verbose:
			describe_xform(descr, fac)
		descr = [ {'id': 'Mixture%d'% i} for i in range(o.shape[1]) ]
		hdr['XFORM'] = XFORM
			
	if PLOT:
		pylab.matshow(Num.transpose(o))
		pylab.show()
	assert o.shape[1] == len(descr), "Length mismatch data=%s descr=%d" % (
						o.shape, len(descr)
						)
	for (i,d) in enumerate(descr):
		hdr['TTYPE%d' % (i+1)] = d['id']
		hdr['F_INFO%d' % (i+1)] = avio.concoct(d)
		hdr['RMS%d' % (i+1)] = math.sqrt(Num.average(o[:,i]**2)) 
	gpkimgclass.gpk_img(hdr, o).write(ofile)





#!/usr/bin/env python


import gpkimgclass

import sys
import os
sys.path.insert(0, os.environ['OXIVOICE'])
import voice_misc


if __name__ == '__main__':
	cutoff_freq = float(sys.argv[1])
	assert cutoff_freq > 0

	d = gpkimgclass.read(sys.argv[2])
	print 'shape=', d.d.shape
	filtered = voice_misc.hipass_sym_butterworth(d.d[:,0], cutoff_freq*d.dt(), order=4)
	gpkimgclass.gpk_img(d.hdr, filtered).write(sys.argv[3])



#!/usr/bin/env python

"""Computes a measure of the peakiness of the waveform.
This is large for pulse trains, and small for hiss-like
signals.
"""


import Num
import math
import die

import sys
import os
sys.path.insert(0, os.environ['OXIVOICE'])
import RMS
import power
import is_voiced
from is_voiced import hipass_sym_butterworth



DT = 0.01	# Seconds.  Default output sampling interval.
HPF = 40.0	# Hz
Final_Smooth_Width = 0.030
SilenceTune = 0.85
DTSMOOTH1 = math.sqrt(Final_Smooth_Width**2 - DT**2/12.0)
DTSMOOTH2 = math.sqrt((SilenceTune*Final_Smooth_Width)**2 - DT**2/12.0)



def peakiness(data, dt, DT, HPF):
	LIM = 10.0
	EPS = 1e-6
	avd = RMS.compute_intensity(data, dt, DT, 'avd', HPF, DTSmooth=DTSMOOTH1)
	aavd = Num.sum(avd)/avd.shape[0]
	rms = RMS.compute_intensity(data, dt, DT, 'rms', HPF, DTSmooth=DTSMOOTH2)
	return Num.minimum(1.0, Num.log(Num.maximum(1.0, rms/(avd + EPS*aavd))))


if __name__ == '__main__':
	import gpkimgclass
	arglist = sys.argv[1:]
	columns = None
	signalfile = None
	outfile = None
	extrahdr = {}
	while len(arglist)>0 and arglist[0][0]=='-':
		arg = arglist.pop(0)
		if arg == '-c':
			if columns == None:
				columns = []
			tmp = arglist.pop(0)
			try:
				col = int(tmp)
			except ValueError:
				col = tmp
			columns.append(col)
		elif arg == '-f':
			signalfile = arglist.pop(0)
			die.note("signalfile", signalfile)
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-dt':
			DT = float(arglist.pop(0))
			assert DT>0.0, 'Silly time step.'
		elif arg == '-hpf':
			HPF = float(arglist.pop(0))
			assert HPF > 0.0, "Silly highpass frequency"
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.die("Unrecognized flag: %s" % arg)
	if len(arglist) > 0:
		rootname = arglist.pop(0)
		if signalfile == None:
			signalfile = rootname + '.d'
		if outfile == None:
			outfile = '%s.pkns.dat' % rootname
	if outfile == None:
		outfile = 'pkns.dat'
	signal = gpkimgclass.read(signalfile)
	nch = signal.d.shape[1]
	if columns == None:
		columns = range(nch)
	die.note("columns", str(columns))
	for c in columns:
		try:
			signal.column(c)
		except KeyError:
			die.info("File has %d columns" % signal.n[1])
			die.die("Bad column: %s" % str(c))

	tmp = [
		peakiness(signal.column(i), signal.dt(), DT, HPF)
			for i in columns
		]
	n = len(tmp[0])
	o = Num.zeros((n,), Num.Float)
	for t in tmp:
		Num.add(o, t, o)
	o = o / nch

	hdr = signal.hdr.copy()
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 0
	hdr['CRVAL2'] = signal.start()
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = 'peakiness'
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	oo = gpkimgclass.gpk_img(hdr, Num.transpose( [o]))
	oo.write(outfile)




"""Viterbi search."""

import Num


HUGE = 1e30


# The internal representation of a path is a linked list of tuples: state: (state_id, prev_state)

def path(nodecost, linkcost, extra_info, N, T):
	"""This function returns a tuple of the cost of the best path
	and the best path, shown as an array: (cost, [j])

	nodecost(t, extra_info) -> ([j-j0], j0, jn)  at time t, pitch=j
	All nodes with j<j0 or j>=jn have infinite cost.
	linkcost(t, extra_info) -> ([j1, j2-j0], j0, jn)  this is the
	cost to go from (t,j1) to (t+1,j2),
	   where j=pitch.  Note the first index spans from 0 to N.
	All links that have either j1 or j1 outside of [j0,jn) have infinite cost.
	The extra_info argument isn't used by this function;
	it is simply passed down to the nodecost() and linkcost
	functions.
	T=number of time steps
	N = Largest possible pitch
	"""

	bestpathto = []
	tmp, j0, jn = nodecost(0, extra_info)
	assert j0>=0 and j0<=jn and jn<=N
	cost = Num.zeros((N,), Num.Float) + HUGE
	cost[j0:jn] = tmp
	bestpathto = [ (j, None) for j in range(N) ]

	for t in range(1,T):
		nbp = []
		ncost = Num.zeros((N,), Num.Float)
		nbp = Num.zeros((N,), Num.PyObject)
		nodecost_t, j0, jn = nodecost(t, extra_info)
		assert j0>=0 and j0<=jn and jn<=N
		linkcost_t, jj0, jjn = linkcost(t, extra_info)
		jjj0 = max(j0, jj0)
		jjjn = min(jn, jjn)
		for j in range(0, jjj0):
			ncost[j] = HUGE
			nbp[j] = None
		for j in range(jjj0, jjjn):
			cc = cost + linkcost_t[:,j-jj0]
			o = Num.argmin(cc)
			ncost[j] = cc[o] + nodecost_t[j-j0]
			nbp[j] = (j, bestpathto[o])
		for j in range(jjjn, N):
			ncost[j] = HUGE
			nbp[j] = None
		bestpathto = nbp
		cost = ncost

	jj = Num.argmin(cost)

	return (cost[jj], _unwind(bestpathto[jj]))


def _unwind(path):
	"""Converts the intermediate nested tuple representation to a linear list."""
	o = []
	while 1:
		id, prev = path
		o.append(id)
		if prev is None:
			o.reverse()
			return o
		path = prev



def _test_nodecost(t, xtra):
	return (xtra[0][t,:], 0, len(xtra[0][t,:]))

def _test_linkcost(t, xtra):
	return (xtra[1], 0, len(xtra[1][0,:]))

def _test1():
	T = 10
	N = 7
	noc = Num.zeros((T,N), Num.Float) + 5
	noc[:,3] = 1
	linkc = Num.zeros((N,N), Num.Float) + 2

	c, p = path(_test_nodecost, _test_linkcost, (noc, linkc), N, T)
	assert abs(c - (10*1 + 9*2)) < 0.0001
	assert p == [3]*T

def _test2():
	T = 10
	N = 7
	noc = Num.zeros((T,N), Num.Float) + 5
	for i in range(T):
		noc[i,i%N] = 1
	linkc = Num.zeros((N,N), Num.Float) + 2

	c, p = path(_test_nodecost, _test_linkcost, (noc, linkc), N, T)
	assert abs(c - (10*1 + 9*2)) < 0.0001
	assert p == [0,1,2,3,4,5,6,0,1,2]


def _test3():
	T = 10
	N = 7
	noc = Num.zeros((T,N), Num.Float) + 1
	noc[0,0] = 0
	linkc = Num.zeros((N,N), Num.Float) + 2
	for i in range(N):
		linkc[i,(i+1)%N] = 1

	c, p = path(_test_nodecost, _test_linkcost, (noc, linkc), N, T)
	assert abs(c - (0+9*1 + 9*1)) < 0.0001
	assert p == [0,1,2,3,4,5,6,0,1,2]

def _test_nodecost3(t, xtra):
	a, j0, jn = _test_nodecost(t, xtra)
	return (a[2:-1], 2, len(a)-1)

def _test4():
	T = 10
	N = 7
	noc = Num.zeros((T,N), Num.Float) + 5
	noc[:,3] = 1
	linkc = Num.zeros((N,N), Num.Float) + 2

	c, p = path(_test_nodecost3, _test_linkcost, (noc, linkc), N, T)
	assert abs(c - (10*1 + 9*2)) < 0.0001
	assert p == [3]*T

def _test_linkcost3(t, xtra):
	a, j0, jn = _test_linkcost(t, xtra)
	return (a[:,2:-1], 2, a.shape[1]-1)

def _test5():
	T = 10
	N = 7
	noc = Num.zeros((T,N), Num.Float) + 5
	noc[:,3] = 1
	linkc = Num.zeros((N,N), Num.Float) + 2
	linkc[3,3] = 1

	c, p = path(_test_nodecost3, _test_linkcost3, (noc, linkc), N, T)
	assert abs(c - (10*1 + 9*2 - (T-1)*1)) < 0.0001
	assert p == [3]*T


if __name__ == '__main__':
	_test1()
	_test2()
	_test3()
	_test4()
	_test5()



#!/usr/bin/env python

"""Perceptual spectrum for speech."""

import Num
import math
import erb_scale
import gpkmisc

# import wavio
# import gpkimgclass

PLOT = True

if PLOT:
	try:
		import pylab
	except ImportError:
		pylab = None
else:
	pylab = None

import sys
import os
sys.path.insert(0, os.environ['OXIVOICE'])
import voice_misc as VM
import power

CBmax = erb_scale.f_to_erb(8000.0)
CBmin = erb_scale.f_to_erb(50.0)
BBSZ = 0.5
E = 0.333
Neural_Tick = 1e-4	# The best possible time resolution of the auditory
			# system.  (E.g. from binaural direction measurements.)
DISS_BWF = 0.6
DISS_FREQ = 7.0

def SafeExp(x):
	return Num.exp(Num.maximum(-200.0, x))


def tau_lp(fc):
	"""This is the cutoff frequency of the modulation transfer function
	in the ear, as a function of frequency.
	From R. Plomp and M. A. Bouman, JASA 31(6), page 749ff, June 1959
		'Relation of hearing threshold and duration of tone pulses'
		"""
	Tau200 = 0.4
	Tau10k = 0.13
	tmp = Tau200 + (math.log(fc/200.0)/math.log(1e4/200.0))*(Tau10k-Tau200)
	assert tmp>0
	return tmp

def cochlear_filter(fofc):
	"""Transfer function taken from "Improved Audio Coding Using a
	Psychoacoustic Model Based on a Cochlear Filter Bank
	IEEE Transactions of Speech and Audio Processing, 10(7) October 2002,
	Pages 495-503.  au=Frank Baumgarte.
	"""
	Q = 4.0
	Slp = 25.0/(20.0*math.log10(1.2))
	Shp = 8.0/(20.0*math.log10(1.2))
	fofc_Shp = fofc**Shp
	return (fofc_Shp) / ( (1+fofc**Slp) * (1 + (1j/Q)*(fofc**(Shp/2))
				- fofc_Shp)
				)


def threshold(f):
	"""IN pressure amplitude.
	Crudely taken from Handbook of Perception vol 4: hearing,
	E.C.Carterette and M.P.Friedman, editors,
	Academic Press 1978, isbn 0-12-161904-4.
	Curve near 70db used."""
	return math.sqrt((200.0/f)**4 + 1 + (f/8000.0)**16)


def accum_abs_sub(a, b, o):
	"""This does a little computation in a block-wise fashion
	to keep all the data witin the processor's cache.
	It computes sum( abs(a-b) ).
	"""
	BKS = 6000
	n = o.shape[0]
	assert a.shape == (n,)
	assert b.shape == (n,)
	assert o.shape == (n,)
	tmp = Num.zeros((BKS,), Num.Float)
	for i in range(0, n, BKS):
		e = min(i+BKS, n)
		tmpi = tmp[:e-i]
		ai = a[i:e]
		bi = b[i:e]
		oi = o[i:e]
		Num.subtract(ai, bi, tmpi)
		Num.absolute(tmpi, tmpi)
		Num.add(oi, tmpi, oi)
	return o


def list_accum_abs_sub_dly(tk, dly):
	len = tk[0][dly:].shape[0]
	o = Num.zeros((len,), Num.Float)
	for tkd in tk:
		a = tkd[:-dly]
		b = tkd[dly:]
		accum_abs_sub(a, b, o)
	return o

	
	

def process_voicing(tk, tick, Dt):
	LOWF = 50.0
	HIGHF = 400.0
	lowf = int(round(1.0/(tick*LOWF)))
	highf = int(round(1.0/(tick*HIGHF)))
	tmps = None
	tmpmin = None
	## tmpss = None
	tmpn = 0
	for dly in range(highf, lowf):
		dhl = dly//2
		dhr = dly - dhl
		# dlytmp = None
		# for tkd in tk:
			# tmp = Num.absolute(tkd[:-dly]-tkd[dly:])
			# if dlytmp is None:
				# dlytmp = tmp
			# else:
				# Num.add(dlytmp, tmp, dlytmp)
		# Num.divide(dlytmp, len(tk), dlytmp)
		dlytmp = list_accum_abs_sub_dly(tk, dly)
		assert len(dlytmp.shape) == 1
		al = Num.zeros((dhl,), Num.Float)
		ar = Num.zeros((dhr,), Num.Float)
		tmp1 = Num.concatenate((al, dlytmp, ar))
		tmp2, t0 = power.smooth(tmp1, tick, Dt, extra=1.0/(LOWF*math.sqrt(12.0)))

		if tmps is None:
			tmpn = 1
			tmps = tmp2
			## tmpss = tmp2**2
			tmpmin = Num.array(tmp2, copy=True)
		else:
			tmpn += 1
			Num.add(tmps, tmp2, tmps)
			## Num.add(tmpss, tmp2**2, tmpss)
			# print 'dly=', dly
			# print 'tmp2=', tmp2[:10]
			# print 'tmpmin=', tmpmin[:10]
			Num.minimum(tmpmin, tmp2, tmpmin)
			# tmpmin = Num.minimum(tmpmin, tmp2)
	## stdev = Num.sqrt(Num.maximum(0.0, (tmpss-tmps**2/tmpn)/(tmpn-1) ))
	## assert Num.alltrue(Num.greater_equal(stdev, 0.0))
	avgmmin = tmps/tmpn - tmpmin
	assert Num.alltrue(Num.greater_equal(avgmmin, 0.0))
	##return (stdev, avgmmin)
	return avgmmin


class dissonance_c(object):
	def __init__(self):
		self.y = None
		self.n = 0
		self.dt = None
	
	def add(self, y, dt, diss_bw):
		if self.dt is None:
			assert dt > 0
			self.dt = dt
		else:
			assert abs(dt-self.dt) < 0.0001*self.dt
		tmp = VM.lowpass_sym_butterworth(y, diss_bw*dt)
		tmp = VM.hipass_first_order(tmp, DISS_FREQ*dt)
		tmp = Num.absolute(tmp)
		# if PLOT and pylab:
			# print 'dissonance_c.add', Num.average(tmp), diss_bw
			# pylab.plot(tmp)
		if self.y is None:
			self.y = tmp
		else:
			Num.add(self.y, tmp, self.y)
		self.n += 1

	def get(self, Dt):
		# if PLOT and pylab:
			# pylab.figure()
		yfilt, t0 = power.smooth(self.y, self.dt, Dt)
		# if PLOT and pylab:
			# pylab.plot(yfilt)
			# pylab.show()
		return yfilt/self.n


def perceptual_spec(data, dt, DT, bmin=CBmin, bmax=CBmax, db=BBSZ, do_mod=False, do_dissonance=False):
	"""This returns something roughly like the neural signals leaving the ear.
	It filters into 1-erb-wide bins, then takes the cube root of
	the amplitude.
	"""
	Neural_Time_Resolution = max(Neural_Tick, dt)
	m = data.shape[0]
	assert m*dt > DT, "Tiny amount of data: %d*%g=%.2gs < DT=%2g" % (m, dt, m*dt, DT)
	ntau = int(round(m*dt/DT))
	d = VM.dataprep_flat_real(data, dt, DT)
	ss = Num.FFT.real_fft( d )
	f = Num.absolute(VM.real_fft_freq(ss.shape[0], d=dt))
	nbands = int(math.ceil((bmax-bmin)/db))
	if do_mod:
		## nbands += 2
		nbands += 1
		tk = []
	if do_dissonance:
		nbands += 1
		diss = dissonance_c()
	neural = Num.zeros((nbands, ntau), Num.Float)
	bctrs = []
	t0x = []
	iband = 0
	b = bmin
	while b < bmax:
		# print 'iband=', iband, 'b=', b, 'nbands=', nbands
		fc = erb_scale.erb_to_f(b)
		# # This choice of sigma is such that the integral of mask df = ebw.
		# # So, the effective width matches the critical bandwidth.
		# sigma = erb_scale.ebw(b)/math.sqrt(2*math.pi)
		# # This shouldn't be a gaussian: the high freq edge is too steep:
		# xferfcn = SafeExp(-(f-fc)**2/(2*sigma**2))

		xferfcn = cochlear_filter(f/fc)
		xferfcn[0] = xferfcn[0].real
		xferfcn[-1] = xferfcn[-1].real
		filtered_sig = Num.FFT.inverse_real_fft( ss * xferfcn )
		hair_cells = Num.maximum(filtered_sig/threshold(fc), 0.0)**(2*E)
		# pylab.plot(hair_cells)
		# pylab.title('hair cells')
		# pylab.figure()
		pwrfilt, t0 = power.smooth(hair_cells, dt, DT)
		# pylab.plot(pwrfilt)
		t0x.append(t0)
		pwrfilt = Num.maximum(VM.lopass_first_order(pwrfilt, DT/tau_lp(fc)), 0.0)
		# pylab.plot(pwrfilt)
		# pylab.show()
		neural[iband,:] = pwrfilt[:ntau]
		bctrs.append({'type':'band', 'smooth': tau_lp(fc), 'fc': fc, 'erb': b,
				'E': E, 'id': 'B%.1f' % b
				})

		if do_mod or do_dissonance:
			tkfilt, t0 = power.smooth(hair_cells, dt, Neural_Time_Resolution)
		if do_dissonance:
			diss_bw = DISS_BWF*erb_scale.ebw(b)
			diss.add(tkfilt, Neural_Time_Resolution, diss_bw)
		if do_mod:
			tk.append(tkfilt)
		b += db
		iband += 1
	if do_mod:
		## stdev, avgmmin = process_voicing(tk, Neural_Time_Resolution, DT)
		avgmmin = process_voicing(tk, Neural_Time_Resolution, DT)

		neural[iband,:] = avgmmin[:ntau]*0.15
		# pylab.plot(avgmmin)
		# pylab.title('voicing')
		# pylab.show()
		bctrs.append({'type':'haspitch', 'id':'haspitch1', 'variant': 1,
				'Fentropy': -math.log(len(tk))
				})
		iband += 1
		## neural[iband,:] = stdev[:ntau]
		## bctrs.append({'type':'haspitch', 'id':'haspitch2', 'variant': 2,
				## 'Fentropy': -math.log(len(tk))
				## })
		## iband += 1
		assert len(bctrs) == iband
	if do_dissonance:
		neural[iband,:] = diss.get(DT)[:ntau]*20.0
		# pylab.plot(diss.get(DT))
		# pylab.title('dissonance')
		# pylab.show()
		bctrs.append({'type':'dissonance', 'id':'dissonance1', 'variant': 1,
				'Fentropy': -math.log(diss.n)
				})
		iband += 1
	assert iband == neural.shape[0]

	return (bctrs, neural, gpkmisc.median(t0x))


__version__ = "$Revision: 1.43 $"
__date__ = "$Date: 2007/03/07 23:54:40 $"


if __name__ == '__main__':
	try:
		import psyco
		psyco.full()
	except ImportError:
		pass
	import gpkimgclass
	fname = sys.argv[1]
	data = gpkimgclass.read(fname)
	DT = 0.01
	bctrs, neural, t0 = perceptual_spec(data.column(0), data.dt(), DT,
					do_mod=True, do_dissonance=True)
	o = Num.transpose(neural)
	if PLOT and pylab:
		pylab.matshow(Num.transpose(o))
		pylab.show()
	hdr = {'CDELT2': DT, 'CRPIX2': 1, 'CRVAL2': data.start(),
		'VERSION': __version__,
		'BITPIX': 0,
		'FNAME': fname,
		'DataSamplingFreq': 1.0/data.dt(),
		'CRPIX1': 1, 'CRVAL1': bctrs[0]['erb'],
		'CDELT1': bctrs[1]['erb']-bctrs[0]['erb']
		}
	for (i,b) in enumerate(bctrs):
		hdr['TTYPE%d' % i] = b['id']
	gpkimgclass.gpk_img(hdr, o).write("-")







"""This works with feature_vec.py.   Incomplete."""

def plot_xform(descr, fac):
	DW = 0.5
	WMAX = 60
	DF = 0.5
	FMAX = 80
	WH = WMAX//2
	import pylab
	nin, nout = fac.shape
	assert len(descr) == nin
	for i in range(nout):
		print 'i=', i
		pylab.figure(i)
		tmp = Num.zeros((WMAX,FMAX), Num.Float)
		norm = math.sqrt(Num.average(fac[:,i]**2))
		nf = fac[:,i]/norm
		for j in range(nin):
			dj = descr[j]
			if '|edge|' in dj['type'] or 'haspitch' in dj['type']:
				continue
			if 'burst' in dj['type']:
				k = lambda delta: [1.0,-.125][abs(delta)<=1]
			elif ' edge' in dj['type']:
				k = lambda delta: [1.0, -1.0][delta<0]
			else:
				k = lambda delta: 1.0

			if dj.has_key('erb') and dj.has_key('width'):
				erbs = [dj['erb']]
			elif dj.has_key('erbs') and dj.has_key('width'):
				erbs = dj['erbs']
			else:
				continue
			nfj = nf[j]/dj['width']
			w0 = WH - int(round(dj['width']/(DW*2.0)))
			we = WH + int(round(dj['width']/(DW*2.0)))
			assert w0>=0 and we<WMAX
			for e in erbs:
				ifreq = int(round(e/DF))
				assert ifreq > 0
				print 'w0:we=', w0, we, 'ifreq=', ifreq, 'tmp.shape', tmp.shape, 'nfj=', nfj
				for w in range(w0, we+1):
					tmp[w,ifreq] += k(w-WH)*nfj
					tmp[w,ifreq-1] += k(w-WH)*nfj/2.0
					tmp[w,ifreq+1] += k(w-WH)*nfj/2.0
		maxabs = gpkmisc.N_maximum(Num.absolute(Num.ravel(tmp)))
		pylab.imshow(Num.transpose(tmp), vmin=-maxabs, vmax=maxabs)
		print



import djt_spectrum
import math
import Num
import voice_misc



def cep(d, n, K, step, w=None, exfac1=1, exfac2=1):
	assert len(d.shape)==1
	if w==None:
		w = 0.5*float(K+2)/float(n)
	ns = int(math.floor((d.shape[0]-n)/step))
	o = []
	for i in range(ns):
		print "dl=", d[i*step:(i+1)*step].shape, "w=", w, "k=", K, "ex1=", exfac1
		q = djt_spectrum.transform(d[i*step:i*step+n], w, K, exfac1)
		c = djt_spectrum.q_cepstrum(q, exfac2)
		o.append(c)
	return Num.array(o, copy=True)


def spec(d, K, w=None, exfac=1, wts=None):
	assert len(d.shape)==1
	n = len(d)
	if w==None:
		w = 0.5*float(K)/float(n)
	print "w=", w, "K=", K, "n=", n
	q = djt_spectrum.transform(d, w, K, exfac)
	c = djt_spectrum.S(q, wts)
	return c

def logspec(d, K, w=None, exfac=1, wts=None):
	return Num.log(spec(d, K, w, exfac, wts))


def stepped_spec(d, n, K, step, w=None, exfac=1, wts=None):
	ns = int(math.floor((d.shape[0]-n)/step))
	o = []
	for i in range(ns):
		c = spec(d[i*step:i*step+n], K, w, exfac, wts)
		o.append(c)
	return Num.array(o, copy=True)



# print factor(1)
# print factor(2)
# print factor(3)
# print factor(4)
# print factor(5)
# print factor(6)
# print factor(7)
# print factor(8)
# print factor(9)
# print factor(10)
# print factor(11)
# print factor(12)


if __name__ == '__main__':
	import gpkimgclass
	K = 4
	tmax = 1.0/40.0	# seconds
	tau_step = 0.010	# seconds.
	tau = 2 * tmax
	if 1:
		a = gpkimgclass.read('foo.d')
		dt = a.dt();
		data = a.d[:,0]
		hdr = a.hdr.copy()
	else:
		hdr = {'CRPIX2':0}
		dt = 1.0/8000.0
		t = Num.arrayrange(4096)*(2*math.pi/50.0)
		data = (Num.exp(5*Num.cos(t))-Num.exp(4*Num.cos(t-0.8)))*Num.cos(t/20)+Num.cos(t*t*13.0)+Num.sin(100.0*Num.sqrt(t))
	od = gpkimgclass.gpk_img(hdr, Num.transpose(Num.array((data,))))
	od.write('tmp.fits')
	nw = voice_misc.near_win_size(tau/dt, 5)
	step = int(round(tau_step/dt))
	print "step=", step, "nw=", nw, "dt=", dt
	os = stepped_spec(data, nw, K, step, w=None, exfac=2)
	oc = cep(data, nw, K, step, w=None, exfac1=2, exfac2=2)
	hdr['CDELT2'] = nw * dt
	hdr['CRPIX2'] = float(hdr['CRPIX2'])/step
	hdr['CRVAL1'] = 0
	hdr['CRPIX1'] = 0
	# hdr['CDELT1'] = dt/2
	hdr['CDELT1'] = math.pi/dt
	oos = gpkimgclass.gpk_img(hdr, os)
	oos.write('bar.fits')
	ooc = gpkimgclass.gpk_img(hdr, oc)
	ooc.write('gleep.fits')



#!python
from setuptools import setup
# from ez_setup import use_setuptools
# use_setuptools()

# from distutils.core import setup, Extension


setup(name = "voicing", version = "0.51",
	description = "Speech signal processing libraries and scripts.",
	author = "Greg Kochanski",
	url = "http://kochanski.org/gpk/code/speechresearch/voicing",
	author_email = "gpk@kochanski.org",
	packages = ['voicing'],
	package_dir = {'voicing': 'lib'},
	scripts = ['emphasis_stevens.py'],
	install_requires = [
				'numpy >= 1.0.3',
				'gmisclib >= 0.51',
				],
	extras_require = {},
	dependency_links = [
				],
	license = 'GPL2',
	keywords = "phonetics speech computational linguistics basic library python science optimize",
	platforms = "All",
	classifiers = [
		'License :: OSI Approved :: GNU General Public License (GPL)',
		'Operating System :: OS Independent',
		'Development Status :: 3 - Alpha',
		'Topic :: Scientific/Engineering',
		'Environment :: Console',
		'Intended Audience :: Science/Research',
		'Programming Language :: Python',
		# X11
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		],
	long_description = """Long
	"""
	)




#!/usr/bin/env python

"""Perceptual loudness measurement.
This is an implementation of
"Percieved Level of Noise by Mark VII and Decibels (E)"
S. S. Stevens,
J. Acoustical Soc. Am. v. 51(2, part 2) 1972
pages 575-602.

The algorithm matches the paper for FractOct==ThirdOct or OneOct.
The algorithm assumes that the level defined by SIXTY_EIGHT
corresponds to 68dB relative to 20 micro-Newtons per square meter
sound pressure.

USER INFORMATION:

You'll find, somewhere, a subdirectory named
speechresearch/voicing .
Within this is a script called emphasis_stevens.py .
This can be run as
python emphasis_stevens.py -o outputfile -write BITPIX=0 -c channelnumber inputfile
(there are a couple of other flags, too).

It will then read the input file (which is an audio recording)
and produce a time-series of the loudness.   Data input and
output is via the gpkio library, in .../speechresearch/gpkio.
That library has many virtues, but reading WAV files is not
one of them, so you have to convert .wav files to
the "GPK ASCII image" format (or one of several other formats),
using  .../speechresearch/lib/wavio.py .

The output will be in an ASCII version of the format,
which should be reasonably intelligible.    (The GPK ASCII image
format is based on the FITS astronomy format from NASA.)

However, that script uses the gpkio and gpklib libraries
(also under ../speechresearch).  These need to be compiled,
and it uses the gpk_img_python package that needs to be
installed via   "python setup.py install".
Oh, and .../speechresearch/gmisclib needs to be in
your PYTHONPATH.

Give it a try, and I'll be happy to help, and will
incorporate the troubles you have into some form of
documentation.    Sorry, I have nothing better yet.
"""

import Num
import math
import Numeric_gpk
import die
import gpkmisc

import sys
import os
sys.path.insert(0, os.environ['OXIVOICE'])
import voice_misc



FAC = 0.5
ThirdOct = 2.0**(1.0/3.0)
OneOct = 2.0
FracOct = 2.0**0.7
Fmax = 12500.0
Fmin = 50.0
NTOBF = int(round(math.log(Fmax/Fmin)/(math.log(FracOct)*FAC)))
EXFAC = 2
E = 0.3333

# 1.0 would cheat on the low frequency end a little.
TAUFAC = 1.2

# Fsum is correct at 68Db SOUND PRESSURE LEVEL AT 3150Hz
Fsum = 0.23*((1+math.log(FracOct)/math.log(ThirdOct))/2.0)

SIXTY_EIGHT = 6000.0
E = 0.3333

def frac_oct_ctr(i):
	return Fmin * (FracOct**i)


def frac_oct_hbw(i):
	return 0.5*(frac_oct_ctr(i+0.5) - frac_oct_ctr(i-0.5))

def calc_ear():
	"""At 68dB SPL."""
	F0 = 50.0
	app_A = [0.729, 1.33, 2.28, 2.92, 3.65, 4.47, 5.40,
		6.40, 7.48, 8.64, 8.64, 8.64, 8.64, 8.64, 8.64,
		10.1, 11.8, 13.7, 16.0, 16.0, 16.0, 16.0, 16.0,
		11.8, 8.64 ]
	o = Num.zeros((NTOBF,), Num.Float)
	for i in range(NTOBF):
		f = frac_oct_ctr(i*FAC)
		idx = math.log(f/F0)/math.log(FracOct)
		iidx = int(round(idx))
		# print i, f, iidx, len(app_A)
		assert iidx>=0 and iidx<len(app_A)
		fidx = idx - iidx
		if abs(fidx) < 0.01:
			o[i] = app_A[iidx]
		elif fidx < 0:
			assert i > 0
			o[i] = app_A[iidx]*(1-abs(fidx)) + app_A[iidx-1]*abs(fidx)
		else:
			assert i+1 < len(app_A)
			o[i] = app_A[iidx]*(1-abs(fidx)) + app_A[iidx+1]*abs(fidx)
	return o


def filter_fcn(f, fc, w):
	Exp = 6	# 18 dB/octave
	return 1.0/(((f-fc)/w)**Exp + 1)


_ff_cache = {}
def cached_filter_fcn(f, fc, w):
	key = (f.shape, fc, w)
	try:
		return _ff_cache[key]
	except KeyError:
		pass
	tmp = filter_fcn(f, fc, w)
	_ff_cache[key] = tmp
	return tmp

	

Ear = calc_ear()

def one_loud(d, extra):
	"""Approximate loudness of the sound in data array d.
	Extra contains misc. parameters."""

	n = d.shape[0]
	dt = extra['dt']
	assert d.shape == (n,)
	ss = Num.FFT.fft(Numeric_gpk.zero_pad_end(d*voice_misc.window(n, 0), EXFAC-1))
	ss[0] = 0	# Remove DC offset.
	# print "# ss.shape=", ss.shape
	f = Num.absolute(voice_misc.fft_freq(ss.shape[0], d=dt))
	assert abs(gpkmisc.N_maximum(f)-0.5/dt) < 0.05/dt
	loud = 0.0
	q = Num.zeros((NTOBF,), Num.Float)
	for i in range( NTOBF ):
		fc = frac_oct_ctr(i*FAC)
		w = frac_oct_hbw(i*FAC)
		q[i] = Num.sum(Num.absolute(ss * filter_fcn(f, fc, w))**2)
		# Fsig is now a bandpass filtered version of d.
	S = Ear * (q**E)
	Smax = S[Num.argmax(S)]
	Ssum = Num.sum(S)/FAC
	loud = Smax + Fsum*(Ssum - Smax)
	return loud




def one_emphasis(t, data, extra, DT):
	"""Measure the loudness near time t."""
	s, e = voice_misc.start_end(t*DT, extra['dt'], extra['window'], data.shape[0])
	d = data[s:e]
	f1 = one_loud(d, extra)
	# print "# t=", t*DT, "s, e=", s, e, "loudness=", f1
	return f1


def emphasis(data, extra, DT=0.01):
	"""Measure the time-series of loudness of a data array data.
	DT is the sampling interval for the resulting loudness time series."""
	ns = int(math.floor(extra['dt']*data.shape[0]/DT))
	o = Num.zeros((ns,), Num.Float)
	for t in range(ns):
		o[t] = one_emphasis(t, data, extra, DT)
	return o


def simple_emphasis(data, dt, DT = 0.01):
	extra = {'dt': dt}
	wmin = 2*frac_oct_hbw(0)
	# Ideally, we'd use a different window size for each frequency...
	taumax = max(TAUFAC / wmin, 1.5*DT)
	extra['window'] = voice_misc.near_win_size(near=taumax/dt, tol=0.1*taumax/dt)
	# print 'win', extra['window'], taumax/dt, wmin, frac_oct_hbw(0), frac_oct_ctr(0)
	return emphasis(data/SIXTY_EIGHT, extra, DT)


if __name__ == '__main__':
	import gpkimgclass
	DT = 0.01	# Seconds.  Default output sampling interval.
	arglist = sys.argv[1:]
	columns = None
	signalfile = None
	outfile = None
	extrahdr = {}
	while len(arglist)>0 and arglist[0][0]=='-':
		arg = arglist.pop(0)
		if arg == '-c':
			if columns == None:
				columns = []
			tmp = arglist.pop(0)
			try:
				col = int(tmp)
			except ValueError:
				col = tmp
			columns.append(col)
		elif arg == '-cal':
			SIXTY_EIGHT = float(arglist.pop(0))
		elif arg == '-f':
			signalfile = arglist.pop(0)
			die.note("signalfile", signalfile)
		elif arg == '-o':
			outfile = arglist.pop(0)
		elif arg == '-dt':
			DT = float(arglist.pop(0))
		elif arg == '-write':
			extrahdr = dict( [q.strip().split('=', 1)
					for q in arglist.pop(0).split(';') ]
					)
		else:
			die.die("Unrecognized flag: %s" % arg)
	if len(arglist) > 0:
		rootname = arglist.pop(0)
		if signalfile == None:
			signalfile = rootname + '.d'
		if outfile == None:
			outfile = rootname + '.emph.dat'
	if outfile == None:
		outfile = 'emph.dat'
	signal = gpkimgclass.read(signalfile)
	nch = signal.d.shape[1]
	if columns == None:
		columns = range(nch)
	die.note("columns", str(columns))
	for c in columns:
		try:
			signal.column(c)
		except KeyError:
			die.info("File has %d columns" % signal.n[1])
			die.die("Bad column: %s" % str(c))

	tmp = [ simple_emphasis(signal.column(i), signal.dt(), DT) for i in columns ]
	n = len(tmp[0])
	o = Num.zeros((n,), Num.Float)
	for t in tmp:
		Num.add(o, t, o)
	o = o / nch

	hdr = signal.hdr.copy()
	hdr['CDELT2'] = DT
	hdr['CRPIX2'] = 0
	hdr['CRVAL2'] = signal.start()
	hdr['CDELT1'] = 1
	hdr['TTYPE1'] = 'loudness'
	hdr['BITPIX'] = -32
	hdr.update( extrahdr )
	oo = gpkimgclass.gpk_img(hdr, Num.transpose( [o]))
	oo.write(outfile)




#!/usr/bin/env python


import math
import Num
import chunkio
import erb_scale
import random
import gpkmisc
import g_pylab as pylab
import die
# import g_threading as GT

import os
import sys
sys.path.insert(0, os.environ['OXIVOICE'])
import feature_vec as FV

DT = 0.005
XFORM = 'xform.chunk'


def get_fv(data, dt, bandonly=False, LengthFac=None, Nsv=None, SpecExp=None,
		DT=DT, simple=False):
	fvf = [FV.feature_vec, FV.simple_feature_vec][simple]
	o, descr, DT, t0 = fvf(data, dt,
					LF=LengthFac, Nsv=Nsv, SpecExp=SpecExp,
					DT=DT)
	print 'pfv: get_fv: DT=', DT
	if bandonly:
		for (i,dtmp) in enumerate(descr):
			if dtmp['type'] == 'haspitch':
				o[i] = Num.zeros(o[i].shape, Num.Float)
	o = Num.transpose(o)
	if XFORM:
		fac = chunkio.datachunk(open(XFORM, 'r')).read_NumArray()
		o = Num.matrixmultiply(o, fac)
	return (o, DT)


def window(t, dur):
	return Num.exp(t**2 * (-0.5/dur**2))


def get_response(erb, dur, ebw, bandonly=True,
			LengthFac=None, SpecExp=None, Nsv=None,
			DT=DT, simple=False):
	Q = 1.21
	dt = 1.0/24000.0
	Ne1 = int(round(0.15/dt))
	Ne2 = int(round(0.15/dt))
	tc = Ne1*dt + 3.0*dur
	Nd = Ne1 + int(round(6*dur/dt)) + Ne2
	sum = None
	M = int(math.ceil( 1.0/(erb_scale.ebw(erb)*dur) ))
	for i in range(M):
		ee = erb + (i-0.5*(M-1))/float(M)
		print 'i=', i, ee
		f = erb_scale.erb_to_f(ee)
		t = Num.arrayrange(Nd)*dt
		phi = t*(2*math.pi*f)
		a = math.cos(Q*i)
		b = math.sin(Q*i)
		d = (a*Num.cos(phi) + b*Num.sin(phi)) * window(t-tc, dur) * erb_scale.ebw(ee)
		# pylab.figure()
		# pylab.title('erb=%.2f i=%d f=%.1f' % (erb, i, f))
		# # pylab.plot(t, d)
		# # pylab.figure()
		o, Dt = get_fv(d, dt, bandonly=bandonly,
				LengthFac=LengthFac, SpecExp=SpecExp, Nsv=Nsv,
				DT=DT, simple=simple)
		# pylab.xlabel('time')
		# pylab.ylabel('feature number')
		# absmax = gpkmisc.N_maximum(Num.absolute(Num.ravel(o)))
		# pylab.imshow(Num.transpose(o), interpolation='nearest', vmax=absmax, vmin=-absmax)
		# # pylab.show()
		print 'DT=', Dt
		assert Dt == DT
		if sum is None:
			sum = o
		else:
			Num.add(sum, o, sum)
	return o/M



def erb_search(erbs, e):
	ebelow = min(erbs)-1
	eabove = max(erbs)+1
	ibelow = None
	iabove = None
	for (i,et) in enumerate(erbs):
		if et<e and et>ebelow:
			ebelow = et
			ibelow = i
		if et>e and et<eabove:
			eabove = et
			iabove = i
	if ibelow is not None and iabove is not None:
		return ibelow + (e-ebelow)*float(iabove-ibelow)/(eabove-ebelow)
	return None



def freq_grid(erbs):
	LFh = [500.0, 1500.0, 2500.0, 3500.0, 4500.0, 5500.0]
	LF = [1000.0, 2000.0, 3000.0, 4000.0, 5000.0, 6000.0]
	for lf in LFh:
		etmp = erb_search(erbs, erb_scale.f_to_erb(lf))
		if etmp is not None:
			pylab.axhline(etmp, color='k', linewidth=1)
	for lf in LF:
		etmp = erb_search(erbs, erb_scale.f_to_erb(lf))
		if etmp is not None:
			pylab.axhline(etmp, color='k', linewidth=2)

def pick_ticks(preferred_num, n):
	delta = int(round(n/preferred_num))
	return range(0, n, delta)


def imshow(img, y_grid=None, x_grid=None,
		nxticks=5, nyticks=5,
		xl=None, xmap=None,
		yl=None, ymap=None,
		yrange = 'data',
		title=None, interpolation='nearest'):

	if yrange == 'symabs':
		maxabs = gpkmisc.N_maximum(Num.absolute(Num.ravel(img)))
		vmin = -maxabs
		vmax = maxabs
	elif yrange == '0+':
		maxval = gpkmisc.N_maximum(Num.ravel(img))
		vmin = 0.0
		vmax = maxval
	elif yrange == 'data':
		vmax = gpkmisc.N_maximum(Num.ravel(img))
		vmin = gpkmisc.N_minimum(Num.ravel(img))

	pylab.imshow(img, vmin=vmin, vmax=vmax, interpolation=interpolation)

	pylab.xlim( 0, img.shape[1])
	pylab.ylim( 0, img.shape[0])
	xticks = pick_ticks(nxticks, img.shape[1])
	pylab.xticks(xticks, [xmap(t) for t in xticks])
	yticks = pick_ticks(nyticks, img.shape[0])
	pylab.yticks(yticks, [ymap(t) for t in yticks])
	if y_grid:
		y_grid()
	if x_grid:
		x_grid()
	if xl:
		pylab.xlabel(xl)
	if yl:
		pylab.ylabel(yl)
	if title:
		pylab.title(title)


def plot(dur, LengthFac=None, SpecExp=None, Nsv=None, DT=DT, simple=False):
	Jc = 15 + int(round(3.0*dur/0.01))
	Ebw = 0.5
	# Ebw = 1.0
	Emax = erb_scale.f_to_erb(6500.0)
	Emin = erb_scale.f_to_erb(60.0)
	# Emax = erb_scale.f_to_erb(1800.0)
	# Emin = erb_scale.f_to_erb(1000.0)
	iErb0 = int(round(Emin/Ebw))
	iErbe = int(round(Emax/Ebw))
	erbs = [ ie*Ebw for ie in range(iErb0, iErbe) ]

	def grw(erb):
		return get_response(erb, dur, ebw=Ebw,
					LengthFac=LengthFac, SpecExp=SpecExp, Nsv=Nsv,
					DT=DT, simple=simple)

	# responses = GT.map(grw, erbs)
	responses = map(grw, erbs)

	# pylab.show()
	assert len(responses) > 0
	assert len(responses[0].shape) == 2
	nf = responses[0].shape[1]
	nt = responses[0].shape[0]
	assert nf <= nt
	ne = len(responses)
	
	# maxabs = 0.0
	# for r in responses:
		# tmp = gpkmisc.N_maximum(Num.absolute(Num.ravel(r)))
		# if tmp > maxabs:
			# maxabs = tmp

	# pylab.gray()
	imsum = Num.zeros((ne,nt), Num.Float)
	for f in range(nf):
		pylab.figure(f)
		img = Num.zeros((ne,nt), Num.Float)
		for ie in range(ne):
			img[ie,:] = responses[ie][:,f]
		if gpkmisc.N_maximum(Num.ravel(img)) < -gpkmisc.N_minimum(Num.ravel(img)):
			Num.multiply(img, -1, img)

		Num.add(imsum, img**2, imsum)

		imshow(img, yrange='symabs', y_grid=lambda: freq_grid(erbs),
				xl='Time (ms)', xmap=lambda j:'%.0f' % (10.0*(j-Jc)),
				yl='Frequency (erbs)', ymap=lambda j: '%.1f'% (Ebw*(j+iErb0)),
				title='Feature %d' % f)
	pylab.figure()
	print_mean_sigma('erbs', Num.sqrt(Num.sum(imsum, axis=1)))
	print_mean_sigma('cs', Num.sqrt(Num.sum(imsum, axis=0)), dur/DT)
	imsum = Num.sqrt(imsum)
	imshow(imsum, yrange='0+', y_grid=lambda: freq_grid(erbs),
			xl='Time (ms)', xmap=lambda j:'%.0f' % (10.0*(j-Jc)),
			yl='Frequency (erbs)', ymap=lambda j: '%.1f'% (Ebw*(j+iErb0)),
			title='Imsum')
	pylab.show()


def print_mean_sigma(label, a, extra=0.0):
	t = Num.arange(a.shape[0])
	avg = Num.sum(t*a)/Num.sum(a)
	sigma = math.sqrt(Num.sum((t-avg)**2 * a)/Num.sum(a) - extra**2)
	print '#MS', label, 'mean=', avg, 'sigma=', sigma


if __name__ == '__main__':
	import sys
	arglist = sys.argv[1:]
	LengthFac = 1.0
	SpecExp = 0.0
	Nsv = 1.00
	Simple = False
	while arglist and arglist[0].startswith('-'):
		arg = arglist.pop(0)
		if arg == '--':
			break
		elif arg == '-SpecExp':
			SpecExp = arglist.pop(0)
		elif arg == '-Nsv':
			Nsv = arglist.pop(0)
		elif arg == '-LF':
			LengthFac = float(arglist.pop(0))
		elif arg == '-sfv':
			Simple = True
		elif arg == '-xform':
			XFORM = arglist.pop(0)
		else:
			die.die('Unrecognized flag: %s' % arg)
	plot(0.02, LengthFac=LengthFac, SpecExp=SpecExp, Nsv=Nsv, simple=Simple)
